"""
Input encoding
"""
import inspect
import numpy as np
import scipy.signal as signal
import copy

from .utils import get_default_kwargs

"""
Shapes of signals

(time,) -> (time, 1)
1D arrays are 1D signals

(time, D)
2D arrays are D-dimensional signals
With D=2 we have a vector signal

(time, H, W)
3D arrays are signals on a WxH grid

(time, H, W, D)
4D arrays are D-dimensional signals on a WxH grid
With D=2 we have a vector signal on a grid

In general signals can take any shape as part of the encoding process.
However, the last step must produce an output of either
(time, 2): a global vector signal
(time, H, W, 2): a local vector signal on a grid

"""

class Encoder:
    """
    Encoder base class

    An encoder translates logical input to an external field protocol.

    Input takes the form of arrays of shape (n_inputs, input_dim).
    1D input arrays may be used as a shorthand for (n_inputs, 1)

    The encoding process consists of one or more steps, where the output of one
    step is input to the next step:

    input -> step1 -> step2 -> ... -> h_ext

    In general, signals can take any shape as part of the encoding process.
    However, the last step must produce an output of either
    (time, 2): a global vector signal
    (time, H, W, 2): a local vector signal on a grid

    Each step is a simple function with optional parameters, e.g.:

    def step(input, param1=default1, param2=default2):
        ...

    The Encoder will inspect the signature of each step to discover the
    available parameters.  The parameters can then be set during encoder
    initialization, or afterwards via `set_params`. Note that parameter names
    may overlap, in which case all matching parameters will be set to the same
    value. """

    """ List of encoder steps. Subclasses must specify this list of steps. """
    steps = None

    def __init__(self, **params):
        self.params = [self._get_default_params(s) for s in self.steps]
        self.set_params(**params)

    def encode(self, input):
        """ Encode input as an external field

        Parameters
        ----------
        input : array_like
            Input shape should be (n_inputs, input_dim).
            1D input arrays may be used as a shorthand for (n_inputs, 1).

        Returns
        -------
        h_ext : ndarray
            The returned array is either:
            1. A global time-dependent field: ``h_ext.shape = (time, 2)``
            2. A spatial time-dependent field: ``h_ext.shape = (time, height, width, 2)``
        """
        input = check_input(input)
        #print(f'input {input.shape}', end=' ')
        for step, params in zip(self.steps, self.params):
            input = step(input, **params)
            #print(f'-> {step.__name__}({params})', end=' ')
            #print(f'-> {input.shape}', end=' ')
            #print(f'-> {input}', end=' ')
        #print('')
        return input

    def __call__(self, input):
        return self.encode(input)

    @classmethod
    def _get_default_params(cls, step):
        return dict(get_default_kwargs(step))

    @classmethod
    def get_default_params(cls):
        params = {}
        for s in cls.steps:
            params.update(cls._get_default_params(s))
        return params

    def get_params(self):
        """ Get available encoder parameters """
        params = {}
        for p in self.params:
            params.update(p)
        return params

    def set_param(self, name, value):
        """ Set encoder parameter name=value """
        for param in self.params:
            if name in param:
                param[name] = value

    def set_params(self, **params):
        """ Set encoder parameters """
        for k, v in params.items():
            self.set_param(k, v)

    def __repr__(self):
        params = {}
        for p in self.params:
            params.update(p)
        params = ", ".join(f"{k}={v}" for k,v in params.items())
        return f"{self.__class__.__name__}({params})"

#
# Encoder steps
#
def expand_dims(input):
    """ Ensure last dimension of input is single-dimensional """
    if input.shape[-1] != 1:
        return np.expand_dims(input, -1)
    return input

def check_input(input):
    input = np.array(input)
    if len(input.shape) == 1:
        return np.expand_dims(input, -1)
    return input

def scale(input, H0=0, H=1):
    """ Scale input to range [H0, H] """
    return H0 + input * (H - H0)

def multiply(input, H=1):
    """ Multiply input by H """
    return H * input

def angle(input, phi0=0, phi=360):
    """ Map input to angle in range [phi0, phi] -> vector components """
    input = check_input(input)
    theta0 = np.deg2rad(phi0)
    theta = np.deg2rad(phi)
    angles = theta0 + input * (theta - theta0)
    # add one dimension for vector
    out = expand_dims(angles)
    out = np.repeat(out, 2, axis=-1)
    out[...,0] = np.cos(out[...,0])
    out[...,1] = np.sin(out[...,1])
    return out

def angle_grid(input, phi0=0, phi=360, grid=None):
    """ Convenience function to offset all angles based on grid """
    input = expand_dims(check_input(input))
    if grid is not None:
        grid = np.atleast_2d(grid)
        phi0 = grid + phi0
        phi = phi0 + phi
    return angle(input, phi0, phi)

def sw(input, rotation=0):
    """ Multiply input vectors by SW astroid """
    input = check_input(input)
    angles = np.arctan2(input[...,1], input[...,0])
    angles = angles - np.deg2rad(rotation)

    Hs = (np.abs(np.sin(angles))**(2/3) + np.abs(np.cos(angles))**(2/3))**(-3/2)
    Hs.shape += (1,)

    return Hs * input

def broadcast_waveform(input, waveform):
    """ Broadcast waveform over input """
    input = check_input(input)
    # multiply input by waveform (broadcasting rules apply)
    out = input * waveform
    # roll last axis to the first (time)
    out = np.moveaxis(out, -1, 1)
    # finally concatenate it
    out = np.concatenate(out)
    return out

def repeat(input, timesteps=1):
    """ Repeat input a number of timesteps """
    return np.repeat(input, timesteps, axis=0)

def sin(input, timesteps=100, phase=0):
    """ Multiply input by sine wave """
    ph = np.deg2rad(phase)
    t = np.linspace(ph, 2 * np.pi + ph, timesteps, endpoint=False)
    waveform = np.sin(t)
    return broadcast_waveform(input, waveform)

def triangle(input, timesteps=100, phase=90):
    """ Multiply input by triangle wave """
    ph = np.deg2rad(phase)
    t = np.linspace(ph, 2 * np.pi + ph, timesteps, endpoint=False)
    waveform = signal.sawtooth(t, 0.5)
    return broadcast_waveform(input, waveform)

def rotate(input, timesteps=100, phase=0):
    """ Multiply input by rotation vector """
    ph = np.deg2rad(phase)
    t = np.linspace(ph, 2 * np.pi + ph, timesteps, endpoint=False)
    cos = broadcast_waveform(input, np.cos(t))
    sin = broadcast_waveform(input, np.sin(t))
    return np.stack([cos, sin], axis=-1)

def ppm(input, pos0=0, pos=1.0, timesteps=100):
    """ Pulse position modulation """
    input = check_input(input)

    # output shape
    shape = input.shape[:-1] if input.shape[-1] == 1 else input.shape
    shape = shape + (timesteps,)

    # calculate location of pulses
    pulse_inds = scale(input, pos0 * timesteps, pos * (timesteps-1)).round().astype(int)
    inds = tuple(np.array(list(np.ndindex(shape[:-1]))).T)
    inds += (pulse_inds.flatten(),)

    # make the pulses
    out = signal.unit_impulse(shape, inds)

    return out

def fixed_vector(input, phi=0):
    """ Convert scalar input to vectors at some angle phi """
    input = check_input(input)
    theta = np.deg2rad(phi)
    # add one dimension to broadcast vector
    input = expand_dims(input)
    vector = [np.cos(theta), np.sin(theta)]
    out = input * vector
    return out

def rotate_vector(input, phi=0):
    """ Rotate vector input by some angle phi """
    input = check_input(input)
    input_shape = input.shape
    theta = np.deg2rad(phi)
    rotation_matrix = np.array(
        [[np.cos(theta), -np.sin(theta)],
         [np.sin(theta),  np.cos(theta)]])
    input = input.reshape((-1,2)).T
    out = np.dot(rotation_matrix, input).T
    out = out.reshape(input_shape)
    return out

def ensure2d(input, fill=1.0):
    """ Convert scalar input to vectors """
    input = check_input(input)
    if input.shape[-1] == 1:
        input = np.column_stack([input, np.full(len(input), fill)])
    return input

# TODO: what's the point of grid_size? Wouldn't it always result in an
# effectively global field?
def grid(input, grid_size=(3,3), grid=None):
    """ Encode input onto grid

    Parameters
    ----------
    grid_size : tuple
        Create grid of fixed size (uniform weights)
        Ignored if grid is set
    grid : array
        Create grid of given size (non-uniform weights)
    """
    if grid is None:
        w, h = grid_size
        grid = np.ones((h, w))
    else:
        grid = np.atleast_2d(grid)

    input = check_input(input)

    # add dimensions to broadcast the grid
    shape = input.shape[:-1] if input.shape[-1] == 1 else input.shape
    input = input.reshape(shape + (1,) * grid.ndim)

    # broadcast the grid on each feature
    out = grid * input

    # now we have:
    # out.shape = (time,) + input.shape + grid.shape
    # shift axes to the end until we have the desired:
    # out.shape = (time,) + grid.shape + input.shape
    for i in range(out.ndim - grid.ndim - 1):
        out = np.moveaxis(out, 1, -1)

    return out

def onehot(input, nbits=2):
    input = check_input(input)

    # roll bit string 00..01 according to input
    out = np.zeros(input.shape + (nbits,))
    out[...,0] = 1
    for i, v in np.ndenumerate(input.astype(int)):
        out[i] = np.roll(out[i], v)

    # roll last axis to the first (time)
    out = np.moveaxis(out, -1, 1)

    # finally concatenate it
    out = np.concatenate(out)

    if input.shape[-1] == 1:
        return out.reshape(out.shape[:-1])
    return out


#
# Encoder definitions
#
class Constant(Encoder):
    """Encode input as the magnitude of a global field.

    Each field value is repeated `timesteps` times.
    The magnitude is scaled between `H0` and `H`.
    The field has a fixed angle `phi`."""

    steps = (repeat, scale, fixed_vector)

class Direct(Encoder):
    """Encode input directly as a global field.

    Note: input should already be in vector form.

    The input vectors are multiplied by `H`.
    """
    steps = (multiply,)

class Angle(Encoder):
    """Encode input as the angle of a global field.

    The angle is scaled between `phi0` and `phi`.
    The field has a fixed magnitude `H`."""

    steps = (angle, multiply)

class Sine(Encoder):
    """Encode input as the amplitude of a sinusoidal global field.

    The amplitude is scaled between `H0` and `H`.
    The field has a fixed angle of `phi` degrees.
    The sine wave has a resolution `timesteps` samples.
    The sine wave has a phase of `phase` degrees.
    """

    steps = (scale, sin, fixed_vector)

class Triangle(Encoder):
    """Encode input as the amplitude of a triangle wave global field.

    The amplitude is scaled between `H0` and `H`.
    The field has a fixed angle of `phi` degrees.
    The triangle wave has a resolution `timesteps` samples.
    The triangle wave has a phase of `phase` degrees.
    """

    steps = (scale, triangle, fixed_vector)

class Rotate(Encoder):
    """Encode input as the amplitude of a rotating global field.

    The amplitude is scaled between `H0` and `H`.
    The rotating field has a resolution `timesteps` samples.
    The rotating field has a phase of `phase` degrees.
    """

    steps = (scale, rotate)

class AngleSine(Encoder):
    """Encode input as the angle of a sinusoidal global field.

    The angle is scaled between `phi0` and `phi`.
    The field has a fixed amplitude `H`.
    The sine wave has a resolution `timesteps` samples.
    The sine wave has a phase of `phase` degrees.
    """

    steps = (angle, expand_dims, sin, multiply)

class AngleTriangle(Encoder):
    """Encode input as the angle of a triangle wave global field.

    The angle is scaled between `phi0` and `phi`.
    The field has a fixed amplitude `H`.
    The triangle wave has a resolution `timesteps` samples.
    The triangle wave has a phase of `phase` degrees.
    """

    steps = (angle, expand_dims, triangle, multiply)

class RotatePulse(Encoder):
    """Encode input as pulses along a rotating global field.

    Each input value is encoded as a pulse in a pulse train of `timesteps`
    samples. The position of the pulse will be between `pos0` and `pos`, where
    0 denotes the beginning of the pulse train and 1 denotes the end.
    The rotating field with have a base magnitude of `H0` and `H` when there is
    a pulse."""

    steps = (ppm, scale, rotate)

class Triangle2D(Encoder):
    """TODO: Please document me"""
    steps = (scale, ensure2d, triangle, rotate_vector)

class OneHot(Encoder):
    """TODO: Please document me"""
    steps = (onehot, scale, triangle, fixed_vector)

class ConstantGrid(Encoder):
    """Encode input as the magnitude of local fields on a grid.

    Each field value is repeated `timesteps` times.

    The grid is defined by `grid`, a 2D array whose values define the weights
    of the input for each grid cell. For example:

    grid = [[0, .5, 0],
            [.5, 1, .5],
            [0, .5, 0]]

    results in a 3x3 grid where the input is applied:
    * unmodified in the center cell (multiplied by 1)
    * halved in the edge cells (multiplied by 0.5)
    * discarded for the corner cells (multiplied by 0)

    After the input has been scaled by the grid weights, it is scaled to the
    range `H0` and `H`. Hence the magnitude of the field in a given cell is
    given by:

    h_ext[i,j] = H0 + grid[i,j] * input * (H - H0)

    For the above example grid, the field:
    * in the center cell will have magnitude H0 + 1 * input * (H - H0)
    * in the edge cells will have magnitude H0 + 0.5 * input * (H - H0)
    * in the corners will have a constant magnitude H0 (regardless of the input)

    Finally, the fields will have a fixed angle `phi`."""

    steps = (repeat, grid, scale, fixed_vector)

class SineGrid(Encoder):
    """Encode input as the amplitude of sinusoidal local fields on a grid.

    The parameters `grid`, `H0`, `H` and `phi` are the same as for `ConstantGrid`.
    The sine wave has a resolution `timesteps` samples and a phase of `phase`
    degrees.
    """
    steps = (grid, scale, expand_dims, sin, fixed_vector)

class TriangleGrid(Encoder):
    """Encode input as the amplitude of a triangle wave local fields on a grid.

    The parameters `grid`, `H0`, `H` and `phi` are the same as for `ConstantGrid`.
    The triangle wave has a resolution `timesteps` samples and a phase of
    `phase` degrees.
    """
    steps = (grid, scale, expand_dims, triangle, fixed_vector)

class AngleSineGrid(Encoder):
    """Encode input as the angle of sinusoidal local fields on a grid.

    The parameters `grid` and `H` are the same as for `ConstantGrid`.
    The grid weights scale the *amplitude* of the sine wave in each
    cell (angle is unaffected).

    The angle is scaled between `phi0` and `phi`.
    The sine wave has a resolution `timesteps` samples and a phase of `phase`
    degrees.
    """
    steps = (angle, expand_dims, sin, multiply, grid) # grid determines H

class AngleGridSine(Encoder):
    """Encode input as the angle of sinusoidal local fields on a grid.

    The parameters `grid` and `H` are the same as for `ConstantGrid`.
    The grid weights are used to offset the angle of each cell.
    The angle is scaled between `phi0` and `phi`.
    The field has a fixed amplitude `H`.
    The sine wave has a resolution `timesteps` samples.
    The sine wave has a phase of `phase` degrees.
    """
    steps = (angle_grid, expand_dims, sin, multiply) # grid adds angle offset

class RotatePulseGrid(Encoder):
    """Encode input as pulses along a rotating field """
    steps = (ppm, grid, scale, rotate)

'''
# TODO: Remove? Or generalize to ASW?
'sw-angle': Encoder(angle, sw, scale),
'sw-rotate': Encoder(scale, rotate, sw),
'sw-pulse': Encoder(ppm, scale, rotate, sw),
'sw-pulse-grid': Encoder(ppm, grid, scale, rotate, sw),
'''



import numpy as np
import pytest
from .utils import approx, assert_allclose
from numpy.testing import assert_array_equal
from itertools import groupby

from flatspin import KagomeSpinIceRotated

def test_init(opencl):
    a = 1/3
    si = KagomeSpinIceRotated(size=(3,3), lattice_spacing=a, **opencl)

    assert si.size == (3,3)
    assert si.spin_count == 6 + 4 + 7 + 4 + 7 + 4 + 6
    assert si.num_neighbors == 4

    si = KagomeSpinIceRotated(size=(2,2), init="ground", **opencl) #check groundstate init
    assert_array_equal(si.spin, [1,  1, -1,  1, -1, -1,  1,  1,  1, -1,  1,  1,  1, -1, -1,  1, -1, 1,  1])

def test_labels():
    # even
    si = KagomeSpinIceRotated(size=(2,2))
    assert_array_equal(si.labels,
            [(0, 0), (0, 1), (0, 2), (0, 3),
             (1, 0), (1, 1), (1, 2),
             (2, 0), (2, 1), (2, 2), (2, 3), (2, 4),
             (3, 0), (3, 1), (3, 2),
             (4, 1), (4, 2), (4, 3), (4, 4)])
    # odd
    si = KagomeSpinIceRotated(size=(3,3))
    assert_array_equal(si.labels,
            [(0, 0), (0, 1), (0, 2), (0, 3), (0, 4), (0, 5),
             (1, 0), (1, 1), (1, 2), (1, 3),
             (2, 0), (2, 1), (2, 2), (2, 3), (2, 4), (2, 5), (2, 6),
             (3, 0), (3, 1), (3, 2), (3, 3),
             (4, 0), (4, 1), (4, 2), (4, 3), (4, 4), (4, 5), (4, 6),
             (5, 0), (5, 1), (5, 2), (5, 3),
             (6, 0), (6, 1), (6, 2), (6, 3), (6, 4), (6, 5)])


def test_indexof():
    si = KagomeSpinIceRotated(size=(4,3))

    inds = [si.indexof(tuple(l)) for l in si.labels]
    assert_array_equal(inds, si.indices())

def test_geometry():
    si = KagomeSpinIceRotated(size=(3,3), lattice_spacing=1/3)
    L = si.L

    # alternating angles
    a60 = np.deg2rad(60)
    assert_allclose(si.angle[L[0::4,0::2]], -a60)
    assert_allclose(si.angle[L[0::4,1::2]], a60)

    assert_allclose(si.angle[L[1::4]], 0)

    assert_allclose(si.angle[L[2::4,0::2]], a60)
    assert_allclose(si.angle[L[2::4,1::2]], -a60)

    # horizontal spacing between "horizontal" magnets is a
    a = si.lattice_spacing
    # vertical spacing between all magnets is a * sqrt(3) / 2
    av = a * np.sqrt(3)/2

    assert_array_equal(si.pos[L[0], 0], 0)
    assert_allclose(si.pos[L[0], 1], a/2 + a*np.arange(6))

    assert_array_equal(si.pos[L[1], 0], av)
    assert_allclose(si.pos[L[1], 1], 2*a*np.arange(4))

    assert_array_equal(si.pos[L[2], 0], 2*av)
    assert_allclose(si.pos[L[2], 1], a/2 + a*np.arange(7))

    assert_array_equal(si.pos[L[3], 0], 3*av)
    assert_allclose(si.pos[L[3], 1], a+2*a*np.arange(4))

    assert_array_equal(si.pos[L[4], 0], 4*av)
    assert_allclose(si.pos[L[4], 1], a/2 + a*np.arange(7))

    assert_array_equal(si.pos[L[5], 0], 5*av)
    assert_allclose(si.pos[L[5], 1], 2*a*np.arange(4))

    assert_array_equal(si.pos[L[6], 0], 6*av)
    assert_allclose(si.pos[L[6], 1], a/2 + a*np.arange(6))

    # test uneven sizes
    si = KagomeSpinIceRotated(size=(1,2), lattice_spacing=1)
    s60 = np.sin(np.deg2rad(60))
    pos = np.array([[0.     , 0.5],
                    [0.     , 1.5],
                    [0.     , 2.5],
                    [0.     , 3.5],
                    [s60    , 0. ],
                    [s60    , 2. ],
                    [s60    , 4. ],
                    [2 * s60, 0.5],
                    [2 * s60, 1.5],
                    [2 * s60, 2.5],
                    [2 * s60, 3.5]])
    a = np.deg2rad(60)
    angle = np.array([-a, a, -a, a, 0, 0, 0, a, -a, a, -a])
    assert_allclose(si.pos, pos)
    assert_allclose(si.angle, angle)

    si = KagomeSpinIceRotated(size=(2,1), lattice_spacing=1)
    pos = np.array([[0.     , 0.5],
                    [0.     , 1.5],
                    [s60    , 0. ],
                    [s60    , 2. ],
                    [2 * s60, 0.5],
                    [2 * s60, 1.5],
                    [2 * s60, 2.5],
                    [3 * s60, 1. ],
                    [3 * s60, 3. ],
                    [4 * s60, 1.5],
                    [4 * s60, 2.5]])
    angle = np.array([-a, a, 0, 0, a, -a, a, 0, 0, a, -a])
    assert_allclose(si.pos, pos)
    assert_allclose(si.angle, angle)
def test_num_neighbors():
    size = (10,10)

    # nearest neighborhood (default)
    si = KagomeSpinIceRotated(size=size)
    assert si.num_neighbors == 4

    # global neighborhood
    si = KagomeSpinIceRotated(size=size, neighbor_distance=np.inf)
    assert si.num_neighbors == si.spin_count - 1


def test_neighbors():
    size = (2,2)
    si = KagomeSpinIceRotated(size=size)
    L = si.L

    # First "horizontal" row
    i = L[0,0]
    ns = list(si.neighbors(i))
    assert ns == [L[0,1], L[1,0]]

    i = L[0,1]
    ns = list(si.neighbors(i))
    assert ns == [L[0,0], L[0,2], L[1,1]]

    i = L[0,3]
    ns = list(si.neighbors(i))
    assert ns == [L[0,2], L[1,2]]

    # First vertical row
    i = L[1,0]
    ns = list(si.neighbors(i))
    assert ns == [L[0,0], L[2,0]]

    i = L[1,1]
    ns = list(si.neighbors(i))
    assert ns == [L[0,1], L[0,2], L[2,1], L[2,2]]

    i = L[1,2]
    ns = list(si.neighbors(i))
    assert ns == [L[0,3], L[2,3], L[2,4]]

    # Middle "horizontal" row
    i = L[2,0]
    ns = list(si.neighbors(i))
    assert ns == [L[1,0], L[2,1], L[3,0]]

    i = L[2,2]
    ns = list(si.neighbors(i))
    assert ns == [L[1,1], L[2,1], L[2,3], L[3,1]]

    i = L[2,4]
    ns = list(si.neighbors(i))
    assert ns == [L[1,2], L[2,3], L[3,2]]

    # Second vertical row
    i = L[3,0]
    ns = list(si.neighbors(i))
    assert ns == [L[2,0], L[2,1], L[4,1]]

    i = L[3,1]
    ns = list(si.neighbors(i))
    assert ns == [L[2,2], L[2,3], L[4,2], L[4,3]]

    i = L[3,2]
    ns = list(si.neighbors(i))
    assert ns == [L[2,4], L[4,4]]

    # Top "horizontal" row
    i = L[4,1]
    ns = list(si.neighbors(i))
    assert ns == [L[3,0], L[4,2]]

    i = L[4,3]
    ns = list(si.neighbors(i))
    assert ns == [L[3,1], L[4,2], L[4,4]]

    i = L[4,4]
    ns = list(si.neighbors(i))
    assert ns == [L[3,2], L[4,3]]

def test_spin_dipolar_field(opencl):
    # default lattice spacing = 1 such that NN islands have coupling strength 1.5
    size = (3,3)
    si = KagomeSpinIceRotated(size=size, **opencl)
    L = si.L

    j_NN = np.array([7/4, np.sqrt(3)/4])
    j_NNN = np.array([-10/(24*np.sqrt(3)), 1/12])

    assert si.spin_dipolar_field(L[0,0], L[0,1]) == approx([-1, 1] * j_NN)
    assert si.spin_dipolar_field(L[0,0], L[1,0]) == approx(j_NN)
    assert si.spin_dipolar_field(L[0,0], L[1,1]) == approx(j_NNN)
    assert si.spin_dipolar_field(L[0,0], L[2,0]) == approx([-1, 1] * j_NNN)

    assert si.spin_dipolar_field(L[1,0], L[0,0]) == approx([1, -1] * j_NN)
    assert si.spin_dipolar_field(L[1,0], L[0,1]) == approx(j_NNN)

    assert si.spin_dipolar_field(L[1,1], L[0,1]) == approx(j_NN)
    assert si.spin_dipolar_field(L[1,1], L[0,2]) == approx([1, -1] * j_NN)
    assert si.spin_dipolar_field(L[1,1], L[2,1]) == approx([1, -1] * j_NN)
    assert si.spin_dipolar_field(L[1,1], L[2,2]) == approx(j_NN)
    assert si.spin_dipolar_field(L[1,1], L[0,0]) == approx([1, -1] * j_NNN)
    assert si.spin_dipolar_field(L[1,1], L[0,3]) == approx(j_NNN)
    assert si.spin_dipolar_field(L[1,1], L[2,0]) == approx(j_NNN)
    assert si.spin_dipolar_field(L[1,1], L[2,3]) == approx([1, -1] * j_NNN)

def test_dipolar_field(opencl):
    size = (3,3)
    si = KagomeSpinIceRotated(size=size, alpha=1.0, neighbor_distance=1, **opencl)
    L = si.L

    j_NN = np.array([7/4, np.sqrt(3)/4])
    j_NNN = np.array([-10/(24*np.sqrt(3)), 1/12])

    assert si.dipolar_field(L[1,1]) == approx([4, 0] * j_NN)
    si.flip(L[2,1])
    assert si.dipolar_field(L[1,1]) == approx(2 * j_NN)
    si.flip(L[0,2])
    assert si.dipolar_field(L[1,1]) == approx([0, 4] * j_NN)

def test_external_field(opencl):
    size = (3,3)
    si = KagomeSpinIceRotated(size=size, neighbor_distance=1, **opencl)
    L = si.L
    a30 = np.deg2rad(30)
    cos30 = np.cos(a30)
    sin30 = np.sin(a30)
    cos60 = np.cos(np.deg2rad(60))
    sin60 = np.sin(np.deg2rad(60))

    assert si.external_field(L[0,0]) == approx([0, 0])
    assert si.external_field(L[1,0]) == approx([0, 0])

    # Horizontal
    si.set_h_ext((1,0))
    assert si.external_field(L[0,0]) == approx([sin30, cos30])
    assert si.external_field(L[0,1]) == approx([sin30, -cos30, ])
    assert si.external_field(L[1,0]) == approx([1, 0])

    # Vertical
    si.set_h_ext((0,1))
    assert si.external_field(L[0,0]) == approx([-cos30, sin30])
    assert si.external_field(L[0,1]) == approx([cos30, sin30])
    assert si.external_field(L[1,0]) == approx([0, 1])

    # 30 degrees
    si.set_h_ext((cos30, sin30))
    assert si.external_field(L[0,0]) == approx([0, 1])
    assert si.external_field(L[0,1]) == approx([sin60, -cos60])
    assert si.external_field(L[1,0]) == approx([cos30, sin30])

    # -30 degrees
    si.set_h_ext((np.cos(-a30), np.sin(-a30)))
    assert si.external_field(L[0,0]) == approx([cos30, sin30])
    assert si.external_field(L[0,1]) == approx([0, -1])
    assert si.external_field(L[1,0]) == approx([sin60, -cos60])

@pytest.mark.parametrize("switching", ["budrikis", "sw"])
def test_flippable(opencl, switching):
    si = KagomeSpinIceRotated(size=(4,4), alpha=1.0, hc=1, switching=switching, **opencl)

    if switching == "budrikis":
        assert len(si.flippable()) == 12
    elif switching == "sw":
        assert len(si.flippable()) == 12

    si.flip(si.L[4,4])
    E = si.switching_energy()
    flippable = [i for i in si.indices() if E[i] > 0]
    assert len(flippable) > 0
    assert_array_equal(si.flippable(), flippable)

    si.polarize()
    si.set_h_ext([-20, 0])
    assert len(si.flippable()) == si.spin_count

def test_find_vertices():
    si = KagomeSpinIceRotated(size=(3,3))
    L = si.L
    vi, vj, indices = si.find_vertices()

    assert_array_equal(vi, [1, 1, 3, 3, 3, 3, 5, 5, 7, 7, 7, 7, 9, 9, 11, 11])
    assert_array_equal(vj, [2, 3, 0, 1, 4, 5, 2, 3, 0, 1, 4, 5, 2, 3, 1, 4])

    assert tuple(indices[0])  == (L[2, 0], L[3, 0], L[2, 1])
    assert tuple(indices[1])  == (L[4, 0], L[3, 0], L[4, 1])

    assert tuple(indices[2])  == (L[0, 1], L[1, 1], L[0, 2])
    assert tuple(indices[3])  == (L[2, 1], L[1, 1], L[2, 2])
    assert tuple(indices[4])  == (L[4, 1], L[5, 1], L[4, 2])
    assert tuple(indices[5])  == (L[6, 1], L[5, 1], L[6, 2])

    assert tuple(indices[6])  == (L[2, 2], L[3, 1], L[2, 3])
    assert tuple(indices[7])  == (L[4, 2], L[3, 1], L[4, 3])

    assert tuple(indices[8])  == (L[0, 3], L[1, 2], L[0, 4])
    assert tuple(indices[9])  == (L[2, 3], L[1, 2], L[2, 4])
    assert tuple(indices[10]) == (L[4, 3], L[5, 2], L[4, 4])
    assert tuple(indices[11]) == (L[6, 3], L[5, 2], L[6, 4])

    assert tuple(indices[12]) == (L[2, 4], L[3, 2], L[2, 5])
    assert tuple(indices[13]) == (L[4, 4], L[3, 2], L[4, 5])

    assert tuple(indices[14]) == (L[2, 5], L[1, 3], L[2, 6])
    assert tuple(indices[15]) == (L[4, 5], L[5, 3], L[4, 6])

def test_vertex_type():
    si = KagomeSpinIceRotated(size=(2,2))
    L = si.L
    # polarized: all vertices type 1

    # make vertex (2,0) type 2 (3 out)
    si.flip(L[2,0])

    # make vertex (2,1) type 2 (3 in)
    si.flip(L[2,3])
    si.flip(L[3,1])

    vertex = si.vertices()

    assert si.vertex_type(vertex[0]) == 1
    assert si.vertex_type(vertex[1]) == 1
    assert si.vertex_type(vertex[2]) == 1
    assert si.vertex_type(vertex[3]) == 1
    assert si.vertex_type(vertex[4]) == 2
    assert si.vertex_type(vertex[5]) == 1

def test_spin_grid(opencl):
    ls = 1/3
    si = KagomeSpinIceRotated(size=(2,2), lattice_spacing=ls, **opencl)

    a30 = np.deg2rad(30)
    a90 = np.deg2rad(90)
    v30n = [-np.sin(-a30), -np.cos(-a30)]
    v30p = [np.sin(a30), np.cos(a30)]
    v90 = [1, 0]
    v0 = [0, 0]

    # grid_spacing == lattice_spacing
    grid = si.spin_grid()
    assert grid.shape == (11, 5, 2)
    expect = [  [v0, v90, v0, v0, v0],
                [v30n, v0, v30p, v0, v0],
                [v0, v0, v0, v90, v0],
                [v30p, v0, v30n, v0, v30p],
                [v0, v90, v0, v0, v0],
                [v30n, v0, v30p, v0, v30n],
                [v0, v0, v0, v90, v0],
                [v30p, v0, v30n, v0, v30p],
                [v0, v90, v0, v0, v0],
                [v0, v0, v30p, v0, v30n],
                [v0, v0, v0, v90, v0]]

    assert_allclose(grid, np.array(expect))



import sys
import numpy as np
import pandas as pd
import itertools
from collections import OrderedDict
from copy import copy
from glob import glob

from .cmdline import param_globals

def parse_spec(spec, ctx=None):
    return eval(spec, param_globals, ctx)

def parse_sweep_spec(specs, ctx={}):
    if not specs:
        return None

    sweep_spec = []

    key, spec = specs[0]

    for value in parse_spec(spec, ctx):
        ctx = copy(ctx)
        ctx[key] = value
        subspec = parse_sweep_spec(specs[1:], ctx)
        if subspec:
            sweep_spec.extend([((key, value),) + sub for sub in subspec])
        else:
            sweep_spec.append(((key, value),))

    return sweep_spec

def parse_repeat_spec(specs, ctx=None):
    repeat_spec = []
    for k, v in specs.items():
        repeat_spec.append([(k, vi) for vi in parse_spec(v, ctx)])

    if not repeat_spec:
        return [()]

    assert all(len(rs) == len(repeat_spec[0]) for rs in repeat_spec)
    return list(zip(*repeat_spec))


def sweep(sweep_spec, repeat=1, repeat_spec=None, params={}, seed=0):
    # NB: order of sweep spec may be important, so need to use OrderedDict
    # TODO: require sweep_spec to be list of tuples instead?
    sweep_spec = list(sweep_spec.items())
    sweep_list = parse_sweep_spec(sweep_spec, ctx=params)

    if not repeat_spec:
        repeat_spec = OrderedDict()
    if repeat > 1:
        repeat_spec['repeat'] = f'range({repeat})'
        repeat_spec['random_seed'] = f'randseed({repeat})'

    for i, sweep_params in enumerate(sweep_list):
        ctx = copy(params)
        ctx.update(sweep_params)

        # Re-seed the RNG to ensure we get the same random sequence across
        # sweeps, but different sequences across repeats
        np.random.seed(seed)

        repeat_list = parse_repeat_spec(repeat_spec, ctx=ctx)
        # print('repeat_list', repeat_list)

        for j, repeat_params in enumerate(repeat_list):
            runparams = dict(sweep_params)
            runparams.update(repeat_params)

            yield i, j, runparams




"""
Command-line related utilities
"""
import sys
import re
import argparse
import fnmatch
import numpy as np
from glob import glob
from collections import OrderedDict
import copy
from tqdm.auto import tqdm
from joblib import Parallel

from flatspin import __version__
from flatspin.data import Dataset
from flatspin.data import read_table

class StoreKeyValue(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        k, v = values.split('=')
        if getattr(namespace, self.dest) is None:
            setattr(namespace, self.dest, OrderedDict())
        d = getattr(namespace, self.dest)
        d[k] = self.parse_value(v)

    def parse_value(self, value):
        return value

class IndexAction(argparse.Action):
    """ index or start:stop:step (all ints) """
    def __call__(self, parser, namespace, values, option_string=None):
        v = parse_index(values)
        setattr(namespace, self.dest, v)

class FilterAction(StoreKeyValue):
    """ key=<filter> where <filter>=start or start:stop (arbitrary types) """
    def parse_value(self, value):
        return parse_filter(value)

class SizeAction(argparse.Action):
    """ <size> or <sizex>x<sizey> """
    def __call__(self, parser, namespace, values, option_string=None):
        v = parse_size(values)
        setattr(namespace, self.dest, v)

class CropAction(argparse.Action):
    """ crop window <crop> or <cropx>,<cropy> where <crop> is before or before:after """
    def __call__(self, parser, namespace, values, option_string=None):
        v = parse_crop(values)
        setattr(namespace, self.dest, v)

class WindowAction(argparse.Action):
    """ window size <sizex>x<sizey> [<stepx>,<stepy>] """
    def __call__(self, parser, namespace, values, option_string=None):
        v = parse_window(values)
        setattr(namespace, self.dest, v)

class DropDuplicatesAction(argparse.Action):
    """column name or 'name1, name2, ..."""
    def __call__(self, parser, namespace, values, option_string=None):
        if getattr(namespace, self.dest) is None:
            setattr(namespace, self.dest, list())
        v = values if values else None
        getattr(namespace, self.dest).append(v)

def parse_index(index_str):
    """ index or start:stop:step (all ints) """
    if ':' in index_str:
        s = index_str.split(':')
        s = list(map(lambda i: int(i) if len(i)>0 else None, s))
        return slice(*s)
    elif '[' in index_str:
        import ast
        return ast.literal_eval(index_str)
    return int(index_str)

def parse_filter(filter_str):
    """ start or start:stop (arbitrary types) """
    if ':' in filter_str and 'lambda' not in filter_str:
        s = filter_str.split(':')
        s = tuple(map(lambda i: parse_filter(i) if len(i)>0 else None, s))
        return slice(*s)
    try:
        return eval(filter_str, {}, {})
    except:
        return filter_str

def parse_size(size):
    """ <size> or <sizex>x<sizey> """
    if 'x' in size:
        return tuple(map(int, size.split('x')))

    return (int(size), int(size))

def parse_crop(crop):
    """ crop window <crop> or <cropx>,<cropy> where <crop> is before or before:after """
    def parse_one_crop(crop):
        if ':' in crop:
            before, after = crop.split(':')
        else:
            before = crop
            after = before

        before = int(before) if before else 0
        after = int(after) if after else 0

        return (before, after)

    if ',' in crop:
        cropx, cropy = tuple(map(parse_one_crop, crop.split(',')))
        return (cropy, cropx)

    crop = parse_one_crop(crop)

    return (crop, crop)

def parse_window(window):
    """ window size <sizex>x<sizey> [<stepx>,<stepy>] """
    try:
        size, step = window
    except ValueError:
        size = window[0]
        step = None

    sizex, sizey = parse_size(size)
    size = (sizey, sizex)

    if step:
        if "," in step:
            stepx, stepy = tuple(map(int, step.split(',')))
            step = (stepy, stepx)
        else:
            step = int(step)
            step = (step, step)
    else:
        step = size

    return size, step

def parse_func(func_str):
    """ fn(arg1, arg2, ...) """
    m = re.match(r'(\w+)\((.*)\)', func_str)
    if m:
        fn = m.group(1)
        c = tuple(map(str.strip, m.group(2).split(',')))
        return (fn,) + c

    return None

def parse_time(time_str, ctx={}):
    """ start:stop:step (arbitrary types, with local ctx) OR index """
    if ":" not in time_str:
        # single index
        return eval(time_str, {}, ctx)

    s = []
    for i in time_str.split(':'):
        if len(i) > 0:
            s.append(eval(i, {}, ctx))
        else:
            s.append(None)
    return slice(*s)

# Functions available to eval_param()
def func_bin(values):
    values = np.array(values, dtype=int)
    max_value = np.max(values)
    assert max_value >= 0, "Values must be unsigned"
    if max_value == 0:
        n_bits = 1
    else:
        n_bits = int(np.floor(np.log2(max_value)) + 1)
    fmt = '{:0' + str(n_bits) + 'b}' # ugh
    bits = list(map(fmt.format, values))
    return bits

def func_randint(low=sys.maxsize, *args, **kwargs):
    return np.random.randint(low, *args, **kwargs)

def func_randseed(n):
    return np.random.randint(2**32, size=n, dtype='int64')

def func_glob(pathname):
    return sorted(glob(pathname))

def func_read_table(table):
    df = read_table(table)
    return df.values


param_globals = {
    'np': np,
    'bin': func_bin,
    'randint': func_randint,
    'randseed': func_randseed,
    'glob': func_glob,
    'inf': np.inf,
    'read_table': func_read_table,
}

def eval_param(v, ctx=None):
    """ Try to eval(v), on failure return v """
    try:
        return eval(v, param_globals, ctx)
    except:
        return v

def eval_params(params, ctx=None):
    d = dict(ctx) if ctx else {}
    for k,v in params.items():
        d[k] = eval_param(v, ctx=d)
    return d

class ProgressBar(tqdm):
    pass

class ParallelProgress(Parallel):
    def __init__(self, progress_bar, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._progress_bar = progress_bar

    def print_progress(self):
        inc = self.n_completed_tasks - self._progress_bar.n
        self._progress_bar.update(inc)

def main_dataset_argparser(description, output_required=False):
    """ Common argparser for scripts which deal with a dataset """
    parser = argparse.ArgumentParser(description=description)

    parser.add_argument('-b', '--datapath', metavar='DATA_PATH',
            help='path to dataset (default: %(default)s)')
    parser.add_argument('-i', '--subset', action=IndexAction, metavar='INDEX',
            help='select a subset of the dataset by index')
    parser.add_argument('-s', '--filter', action=FilterAction, default={},
            help='filter a subset of the dataset by parameter')
    parser.add_argument('-l', '--list', action='store_true',
            help='list the dataset')
    parser.add_argument('-o', '--output', required=output_required,
            metavar='FILE', help='save result(s) to file(s)')
    parser.add_argument('-d', '--drop-duplicates', action=DropDuplicatesAction, nargs="*",
            help='drop duplicates, <column name> or "<column name 1>, ... ,<column name n>"')
    parser.add_argument('-V', '--version', help='display flatspin version', action='store_true')

    return parser

def main_dataset_grid_argparser(description, output_required=False):
    """ Common argparser for scripts dealing with grid operations on a dataset """
    parser = main_dataset_argparser(description, output_required)

    parser.add_argument('-t', default='::spp',
            help='time range start:stop:step (default: %(default)s)')
    parser.add_argument('-g', '--grid', action=SizeAction,
            help='grid size <size> or <sizex>x<sizey> (set GRID to 0 to disable grid)')
    parser.add_argument('-c', '--crop', action=CropAction,
            help='''crop window CROP or CROPx,CROPy where CROP is before or before:after
                    (when GRID is 0, CROP,CROPx,CROPy denote the percentage to crop)''')
    parser.add_argument('-w', '--window', action=WindowAction, nargs='+',
            help='window size <sizex>x<sizey> [<stepx>,<stepy>]')

    return parser

def main_version(args):
    print(f'flatspin version {__version__}')

def main_dataset(args):
    """ Common main function for scripts which deal with a dataset """
    if args.version:
        main_version(args)
        sys.exit(0)

    datapath = args.datapath if args.datapath else '.'
    ds = Dataset.read(datapath)

    if args.filter:
        kwargs = args.filter
        ds = ds.filter(**kwargs)

    if args.subset is not None:
        if type(args.subset) == list:
            ds = ds[args.subset]
        else:
            key = args.subset
            ds = ds[key]

    if args.drop_duplicates is not None:
        for drop_dup in args.drop_duplicates:
            ds = ds.drop_duplicates(subset=drop_dup)

    if args.list:
        print(ds)
        sys.exit(0)

    return ds




from functools import partial
import pytest
import numpy.testing

# TODO: increase opencl precision?
approx = partial(pytest.approx, abs=1e-6)
assert_allclose = partial(numpy.testing.assert_allclose, atol=1e-6)



"""
Utilities related to the flatspin model
"""
import numpy as np
import copy
import itertools
import os
import inspect
from datetime import timedelta
from collections import OrderedDict
import importlib

def numeric_dict(d):
    return {k: int(v) for k,v in d.items() if str(v).isnumeric()}

def filter_dict(d, keys):
    return d.__class__((k,v) for k,v in d.items() if k in keys)

def pop_dict(d, keys):
    return d.__class__((k, d.pop(k)) for k in list(d.keys()) if k in keys)

def try_eval(s):
    """ Try to eval(s), on failure return s """
    try:
        return eval(s, {'array': np.array, 'inf': np.inf})
    except:
        return s

def eval_dict(d):
    return d.__class__((k, try_eval(v)) for k,v in d.items())

def make_tuple(tup):
    if isinstance(tup, tuple):
        return tup
    return (tup,)

def unique_list(l):
    u = []
    for i in l:
        if not i in u:
            u.append(i)
    return u

def get_default_args(func, exclude=['self']):
    sig = inspect.signature(func)
    params = OrderedDict((k,p.default) for k,p in sig.parameters.items() if k not in exclude)
    return params

def get_default_kwargs(func, exclude=[]):
    params = get_default_args(func, exclude)
    params = OrderedDict((k,d) for k,d in params.items() if d is not inspect._empty)
    return params

def get_default_params(cls):
    try:
        default_params = cls.get_default_params()
    except AttributeError:
        if inspect.isclass(cls):
            # Walk up the class hierarchy to discover all available parameters
            # Assumes each sub class' __init__() will only contain new parameters
            # or modified defaults.  Old parameters are delegatet to the parent
            # class through **kwargs
            default_params = {}
            mro = inspect.getmro(cls)
            for cls in reversed(mro):
                default_params.update(get_default_kwargs(cls.__init__))
        else:
            default_params = get_default_kwargs(cls)
    return dict(default_params)

def check_params(params, classes):
    cls_params = split_params(params, classes)
    unknown_params = cls_params[-1]
    if unknown_params:
        raise TypeError("Invalid params", list(unknown_params.keys()))

def filter_params(cls, params):
    valid = get_default_params(cls)
    return filter_dict(params, valid)

def pop_params(cls, params):
    valid = get_default_params(cls)
    return pop_dict(params, valid)

def split_params(params, classes):
    """ Split params belonging to the given list of classes.

    Returns a list of parameters matching the the given list of classes. The
    last element of the list are parameters which didn't match any of the
    class signatures. """

    params = copy.deepcopy(params)
    valid_params = list(map(get_default_params, classes))

    # Check that classes don't have overlapping parameters
    overlap = set.intersection(*[set(p.keys()) for p in valid_params])
    assert not overlap, f"Overlapping parameters for {classes}: {overlap}"

    result = []
    for cls, valid in zip(classes, valid_params):
        cls_params = filter_dict(params, valid)
        result.append(cls_params)
        for k in cls_params.keys():
            del params[k]

    result.append(params) # remaining params

    return result

def format_label(label):
    if isinstance(label, np.ndarray):
        label = tuple(label)
    return str(label).replace(' ', '')

def label_columns(labels, prefix='spin'):
    labels = list(map(format_label, labels))
    return [prefix + str(c) for c in labels]

def label_columns_2d(labels, prefix='spin'):
    cols = label_columns(labels, prefix)
    cols = [[col + 'x', col + 'y'] for col in cols]
    return list(itertools.chain(*cols))

def get_outdir(script_filename):
    root, ext = os.path.splitext(script_filename)
    return root + ".out"

def make_outdir(script_filename):
    outdir = get_outdir(script_filename)
    os.makedirs(outdir, exist_ok=True)
    return outdir

def import_class(name, module=None):
    if '.' in name:
        # Full module path
        module, cls = name.rsplit('.', maxsplit=1)
    else:
        assert module
        cls = name

    module = importlib.import_module(module)
    return getattr(module, cls)

def get_subclasses(module, superclass):
    predicate = lambda cls: inspect.isclass(cls) and issubclass(cls,
            superclass) and cls is not superclass

    return inspect.getmembers(module, predicate)



import numpy as np
import pytest
import time

from flatspin import SquareSpinIceClosed

def test_init(benchmark, opencl):
    size = (10,10)

    def init():
        model = SquareSpinIceClosed(size=size, neighbor_distance=10, **opencl)
        model.dipolar_fields() # Ensure h_dip_cache initialized

    benchmark(init)

@pytest.mark.parametrize("neighbor_distance", [1,5,10,20])
def test_neighbor_distance(benchmark, opencl, neighbor_distance):
    size = (20,20)

    init = lambda: SquareSpinIceClosed(size=size, neighbor_distance=neighbor_distance, **opencl)

    benchmark(init)

def test_dipolar_fields(benchmark, opencl):
    size = (100,100)
    si = SquareSpinIceClosed(size=size, **opencl)
    si.randomize()

    benchmark(si.dipolar_fields)

def test_external_fields(benchmark, opencl):
    size = (100,100)
    si = SquareSpinIceClosed(size=size, **opencl)
    si.randomize()

    benchmark(si.external_fields)

def test_total_fields(benchmark, opencl):
    size = (100,100)
    si = SquareSpinIceClosed(size=size, **opencl)
    si.randomize()

    benchmark(si.total_fields)

def test_flippable(benchmark, opencl):
    size = (100,100)
    si = SquareSpinIceClosed(size=size, hc=0.04, **opencl)
    si.randomize()

    benchmark(si.flippable)

def test_step(benchmark, opencl):
    size = (100,100)
    si = SquareSpinIceClosed(size=size, hc=0.04, **opencl)

    def step():
        si.randomize()

    benchmark(step)

def test_relax(benchmark, opencl):
    size = (50,50)
    si = SquareSpinIceClosed(size=size, hc=0.04, **opencl)

    def relax():
        si.randomize()
        si.relax()

    benchmark(relax)

def test_energy(benchmark, opencl):
    size = (100,100)
    si = SquareSpinIceClosed(size=size, **opencl)
    si.randomize()

    benchmark(si.energy)

def test_total_energy(benchmark, opencl):
    size = (100,100)
    si = SquareSpinIceClosed(size=size, **opencl)
    si.randomize()

    benchmark(si.total_energy)

def test_total_magnetization(benchmark, opencl):
    size = (100,100)
    si = SquareSpinIceClosed(size=size, **opencl)
    si.randomize()

    benchmark(si.total_magnetization)

def test_vertices(benchmark, opencl):
    size = (100,100)
    si = SquareSpinIceClosed(size=size, **opencl)
    si.randomize()

    vertices = lambda: list(si.vertices())

    benchmark(vertices)

def test_vertex_count(benchmark, opencl):
    size = (100,100)
    si = SquareSpinIceClosed(size=size, **opencl)
    si.randomize()

    benchmark(si.vertex_count)

def test_vertex_population(benchmark, opencl):
    size = (100,100)
    si = SquareSpinIceClosed(size=size, **opencl)
    si.randomize()

    benchmark(si.vertex_population)

def test_spin_grid(benchmark):
    size = (100,100)
    si = SquareSpinIceClosed(size=size)
    si.randomize()

    benchmark(si.spin_grid)

def test_update_thermal_noise(benchmark):
    size = (25, 25)
    si = SquareSpinIceClosed(size=size, temperature=300)
    si.randomize()

    benchmark(si.update_thermal_noise)

def test_set_angle(benchmark, opencl):
    size = (10,10)
    model = SquareSpinIceClosed(size=size, neighbor_distance=10, **opencl)
    model.dipolar_fields() # Ensure h_dip_cache initialized
    angle = np.random.uniform(0, 360, model.spin_count)
    angle = np.deg2rad(angle)

    def set_angle():
        model.set_angle(angle)
        model.dipolar_fields() # Ensure h_dip_cache initialized

    benchmark(set_angle)

def test_set_pos(benchmark, opencl):
    size = (10,10)
    model = SquareSpinIceClosed(size=size, neighbor_distance=10, **opencl)
    model.dipolar_fields() # Ensure h_dip_cache initialized
    pos = model.pos + .3*np.random.uniform(-1, 1, model.pos.shape)

    def set_pos():
        model.set_pos(pos)
        model.dipolar_fields() # Ensure h_dip_cache initialized

    benchmark(set_pos)

def test_set_geometry(benchmark, opencl):
    size = (10,10)
    model = SquareSpinIceClosed(size=size, neighbor_distance=10, **opencl)
    model.dipolar_fields() # Ensure h_dip_cache initialized
    pos = model.pos + .3*np.random.uniform(-1, 1, model.pos.shape)
    angle = np.random.uniform(0, 360, model.spin_count)
    angle = np.deg2rad(angle)

    def set_geometry():
        model.set_geometry(pos, angle)
        model.dipolar_fields() # Ensure h_dip_cache initialized

    benchmark(set_geometry)



"""
Dipole model runner
"""
import sys
import os
import numpy as np
import pandas as pd
import time
from numbers import Number
from datetime import timedelta
import subprocess
from copy import copy
from importlib import import_module
import warnings

from .cmdline import ProgressBar
from .data import read_table, to_table, save_table, is_archive, archive_key
from .utils import (pop_params, label_columns, label_columns_2d, eval_dict,
        import_class)

#
# List of deprecated parameters
#
# If you rename a parameter, please add it here to raise an error if it's being
# used! Otherwise it will go unnoticed since we ignore unknown parameters to
# facilitate fancy sweeps such as -s 'sz=arange(10,51,10)' -s 'size=[(sz,sz)]'
#
deprecated_params = ('sw_a', 'sw_alpha', 'temp_interp', 'temp_func', 'temp')

class DeprecationError(Exception):
    pass

run_sample_callback = None

def run(model, encoder='Sine', input=1, input_key=None, periods=100, spp=100,
        **params):
    """
    Run a flatspin simulation

    Parameters
    ----------
    model : SpinIce
        Model instance
    encoder : string or Encoder
        Name of input encoder to use or Encoder instance.
        See flatspin.encoder for a list of available encoders
    input : float, array or string
        Input to be encoded
        float: constant input repeated periods times
        array: array of input to encode
        string: filename of table input data
    input_key : int or str
        Index into input (table column)
    periods : int
        Number of periods of the external field to run
        Only used if input is a float
    spp : int
        Number of samples to save per input value
    params : dict
        Params to pass to the encoder
    """
    # unknown params are ignored

    if isinstance(encoder, str):
        encoder_class = import_class(encoder, 'flatspin.encoder')
        encoder = encoder_class()

    encoder_params = pop_params(encoder, params)
    encoder.set_params(**encoder_params)

    # Bail out if user tries any deprecated params
    for p in params.keys():
        if p in deprecated_params:
            raise DeprecationError("Parameter is deprecated", p)

    # Extract any time-varying parameters
    params_t = {k[:-2]: params.pop(k) for k in list(params.keys()) if k.endswith('_t')}

    for k in params_t:
        set_fn = f"set_{k}"
        if not hasattr(model, set_fn):
            raise ValueError(f"Invalid parameter: {k}_t")

    if "h_ext" in params_t:
        raise ValueError(f"h_ext_t is not supported, use the input encoders to set h_ext")

    # Warn on unknown params
    if params:
        unknown_params = ", ".join(params.keys())
        warnings.warn("Ignoring unknown parameters: " + unknown_params)

    result = {}

    if isinstance(input, str):
        # load input from file
        input = read_table(input)
        result['input'] = input
    elif isinstance(input, Number):
        # constant
        input = input * np.ones(periods)
        result['input'] = pd.DataFrame(input, columns=['input'])
    else:
        # array/list
        input = np.array(input)
        if input.ndim > 1:
            cols = label_columns(np.ndindex(*input.shape[1:]), prefix='input')
            input_flat = input.reshape((input.shape[0], -1))
            result['input'] = pd.DataFrame(input_flat, columns=cols)
        else:
            result['input'] = pd.DataFrame(input, columns=['input'])

    if input_key is not None:
        input = input[input_key]
    input = np.array(input, dtype=float)

    t0 = time.time()

    # Encode input as h_ext
    h_ext = encoder(input)

    # Figure out timesteps per period
    ts = len(h_ext) / len(input)
    if ts % spp != 0:
        print(f"WARNING: spp {spp} does not evenly divide the encoded input")
        print(f"         spp should divide the number of timesteps per input {ts}, e.g. spp={ts}")
    sample_every = int(np.round(ts / spp))

    # Input has now been encoded into h_ext, resulting in spp samples per input
    # value. Repeat input values spp times so that h_ext and input DataFrames
    # can be easily joined on the same index when needed. We keep the original
    # input index as an extra column called "index".
    input_df = result['input']
    input_df = input_df.loc[input_df.index.repeat(spp)].reset_index()
    input_df.index.name = 't'
    result['input'] = input_df

    # Time-varying parameters
    for k, v in list(params_t.items()):
        if isinstance(v, str):
            # load values from file
            v = read_table(v)
            v = np.array(v)
            if v.shape[-1] == 1:
                # Treat single-column tables 1d
                v = v.reshape(v.shape[:-1])

        if len(v) != len(input):
            raise ValueError(f"Length of {k}_t (={len(v)}) does not match length of input (={len(input)})")
        repeats = len(h_ext) / len(v)
        params_t[k] = np.repeat(v, repeats, axis=0)

    # Save h_therm if temperature > 0
    save_h_therm = model.temperature != 0 or \
            ('temperature' in params_t and np.any(params_t['temperature'] != 0))

    result['geometry'] = pd.DataFrame({
        'posx': model.pos[:,0].copy(),
        'posy': model.pos[:,1].copy(),
        'angle': model.angle.copy()})

    result['hc'] = pd.DataFrame({'threshold': model.threshold.copy()})

    cols = label_columns(model.labels, prefix='init')
    result['init'] = pd.DataFrame([model.spin.copy()], columns=cols)

    # TODO: dump model?
    # make a copy of model so we store the initial state in the dataset
    #    'model': copy.deepcopy(model),

    steps = 0

    sample_index = []
    result['spin'] = []
    result['mag'] = []
    result['h_ext'] = []
    if save_h_therm:
        result['h_therm'] = []
    result['energy'] = []
    result['steps'] = []
    if params_t:
        result['params_t'] = []

    msg = type(model).__name__
    disable_progress = not sys.stdout.isatty()
    progress_bar = ProgressBar(desc=msg, total=len(h_ext), disable=disable_progress)

    for i, h in enumerate(h_ext):
        if h.ndim > 1:
            # grid
            model.set_grid('h_ext', h)
        else:
            # global field
            model.set_h_ext(h)

        # Time-varying parameters
        for k, v in params_t.items():
            set_fn = getattr(model, f"set_{k}")
            set_fn(v[i])

        should_sample = (i % sample_every == 0)

        steps += model._relax(copy_gpu_to_cpu=should_sample)

        progress_bar.update()

        if should_sample:
            sample_index.append(i)
            result['spin'].append(model.spin.copy())
            result['mag'].append(model.vectors.flatten())
            result['energy'].append(model.total_energies())
            if save_h_therm:
                if isinstance(model.h_therm, np.ndarray):
                    result['h_therm'].append(model.h_therm.flatten())
                else:
                    result['h_therm'].append(np.zeros(model.spin_count * 2))
            if h.ndim > 1:
                # local field
                result['h_ext'].append(model.h_ext.flatten())
            else:
                # global field
                result['h_ext'].append(h)

            # Time-varying parameters
            if params_t:
                result['params_t'].append([v[i] for v in params_t.values()])

            result['steps'].append(steps)

            if run_sample_callback:
                run_sample_callback(model, i, len(h_ext), result, params)


    progress_bar.close()

    cols = label_columns(model.labels, prefix='spin')
    result['spin'] = pd.DataFrame(result['spin'], columns=cols, index=sample_index, dtype=result['spin'][0].dtype)
    result['spin'].index.name = 't'

    cols = label_columns_2d(model.labels, prefix='mag')
    result['mag'] = pd.DataFrame(result['mag'], columns=cols, index=sample_index, dtype=result['mag'][0].dtype)
    result['mag'].index.name = 't'

    cols = ['h_extx', 'h_exty']
    if h_ext.ndim > 2:
        # local field
        cols = label_columns_2d(model.labels, prefix='h_ext')
    result['h_ext'] = pd.DataFrame(result['h_ext'], columns=cols, index=sample_index, dtype=model.h_ext.dtype)
    result['h_ext'].index.name = 't'

    if save_h_therm:
        cols = label_columns_2d(model.labels, prefix='h_therm')
        result['h_therm'] = pd.DataFrame(result['h_therm'], columns=cols, index=sample_index, dtype=np.float64)
        result['h_therm'].index.name = 't'

    cols = ["E_dip", "E_ext", "E_therm", "E_tot"]
    result['energy'] = pd.DataFrame(result['energy'], columns=cols, index=sample_index, dtype=np.float64)
    result['energy'].index.name = 't'

    result['steps'] = pd.DataFrame(result['steps'], columns=['steps'], index=sample_index, dtype=int)
    result['steps'].index.name = 't'


    runtime = time.time() - t0

    result['stats'] = {
        'runtime': runtime,
        'steps': steps,
    }

    # Time-varying parameters
    if params_t:
        result['params_t'] = pd.DataFrame(result['params_t'], columns=params_t.keys(), index=sample_index)
        result['params_t'].index.name = 't'

    td = timedelta(seconds=runtime)
    print(f'Completed {steps} steps in {td}')

    return result

def run_and_save(model_class, params, outdir, data_format):
    params_table = to_table(params)

    model_params = pop_params(model_class, params)
    model = model_class(**model_params)

    results = run(model, **params)
    results["params"] = params_table

    if not is_archive(outdir):
        os.makedirs(outdir)

    for name, data in results.items():
        df = to_table(data)
        if not is_archive(outdir):
            name = f"{name}.{data_format}"
        save_table(df, os.path.join(outdir, name))

def run_local(dataset, verbose=True):
    """ Run on localhost """
    n_runs = len(dataset)

    mod, cls = dataset.info['model'].rsplit('.', 1)
    module = import_module(mod)
    model_class = getattr(module, cls)
    data_format = dataset.info['data_format']

    for i, ds in enumerate(dataset):
        params = copy(ds.params)
        params.update(eval_dict(ds.row().to_dict()))
        outdir = params['outdir']
        outpath = os.path.join(dataset.basepath, outdir)
        if verbose:
            description = ["{}={}".format(k,v) for (k,v) in ds.row().items()]
            description = " ".join(description)
            print(f"Run {i+1}/{n_runs}: {description}")
        run_and_save(model_class, params, outpath, data_format)

def generate_script(template, outfile, **params):
    with open(template) as fp:
        tpl = fp.read()
    script = tpl.format(**params)
    with open(outfile, 'w') as fp:
        fp.write(script)

job_script_template = os.path.join(os.path.dirname(__file__), 'flatspin.slurm.sh')

def run_dist(dataset, wait=True, max_jobs=1000):
    """ Run distributed on a cluster """

    #
    # Generate job script
    #

    # Construct a sensible name for the job script
    job_script_dir = dataset.basepath
    job_script_name = os.path.basename(job_script_template)
    job_script = os.path.join(job_script_dir, job_script_name)

    # Job template params
    job_params = {
        'job_script_dir': job_script_dir,
        'job_script_name': job_script_name,
        'basepath': dataset.basepath,
    }

    generate_script(job_script_template, job_script, **job_params)

    #
    # Submit jobs
    #

    # array size will never exceed max_jobs
    array_size = min(max_jobs, len(dataset))

    cmd = ['sbatch']
    if wait:
        cmd.append('--wait')
    cmd.append(f'--array=0-{array_size-1}')
    cmd.append(job_script)
    print(cmd)
    p = subprocess.Popen(cmd)

    if wait:
        p.wait()




import pytest

# ugh...
from os.path import abspath, join, dirname
import sys
package_path = abspath(dirname(dirname(__file__)))
sys.path.insert(0, package_path)

def pytest_addoption(parser):
    parser.addoption("--opencl-platform", type=int, default=0)
    parser.addoption("--opencl-device", type=int, default=0)

# Add CUDA fixture if pycuda is installed
try:
    import pycuda
    params=[{'use_opencl': False}, {'use_opencl': True}, {'use_cuda': True}]
    ids=['default', 'opencl', 'cuda']
except ImportError:
    params=[{'use_opencl': False}, {'use_opencl': True}]
    ids=['default', 'opencl']

@pytest.fixture(params=params, ids=ids)
def opencl(request):
    args = request.param
    args['opencl_platform'] = request.config.getoption("--opencl-platform")
    args['opencl_device'] = request.config.getoption("--opencl-device")
    return args



import numpy as np
from numpy.linalg import norm
import pytest
from .utils import approx

from flatspin import IsingSpinIce
from numpy.testing import assert_array_equal, assert_allclose

def ASW(h_par, h_perp, b=1, beta=3, c=1, gamma=3):
    sw = b*(1 - ((h_perp/c)**2)**(1/beta))**(gamma/2)
    sw[h_par<0] *= -1
    return sw

def test_init_no_temp(opencl):
    si = IsingSpinIce(temperature=0, **opencl)

    assert si.h_therm.shape == (si.spin_count, 2)
    assert si.h_therm == approx(0)

    tf = si.thermal_fields()
    assert tf.shape == (si.spin_count, 2)
    assert tf == approx(0)

def test_init_temp(opencl):
    si = IsingSpinIce(temperature=100, **opencl)

    assert si.h_therm.shape == (si.spin_count, 2)
    assert not si.h_therm == approx(0)

    tf = si.thermal_fields()
    assert tf.shape == (si.spin_count, 2)
    assert not tf == approx(0)

    # thermal_fields is in local per-spin reference, while h_therm is global
    assert np.all(tf != si.h_therm)

def test_flip(opencl):
    si = IsingSpinIce(temperature=100, **opencl)

    h_therm1 = si.h_therm.copy()
    tf1 = si.thermal_fields()

    si.flip(0)

    # h_therm unchanged
    assert_array_equal(si.h_therm, h_therm1)

    # thermal_fields for spin 0 should have flipped
    tf1[0] *= -1
    assert_array_equal(si.thermal_fields(), tf1)

def test_relax(opencl):
    si = IsingSpinIce(temperature=300, **opencl)
    tf1 = si.thermal_fields()
    si.relax()

    # relax() updates thermal fields
    assert si.thermal_fields() != approx(tf1)


def test_set_temperature(opencl):
    si = IsingSpinIce(alpha=0,
            hc=0.02, sw_b=1, sw_c=1, sw_beta=3, sw_gamma=3,
            m_therm=860e3 * 1.5e-23, therm_timescale=1, **opencl)

    assert si.temperature == 0
    si.set_temperature(400)
    assert si.temperature == 400

    steps1 = si.relax()
    h_therm1 = si.h_therm.copy()
    assert steps1 > 0

    old_temperature = si.temperature
    si.set_temperature(500)
    assert si.temperature > old_temperature

    steps2 = si.relax()
    h_therm2 = si.h_therm.copy()
    assert steps2 > 0
    assert steps2 > steps1
    assert not np.array_equal(h_therm1, h_therm2)

@pytest.mark.parametrize("temperature", [0, 100, 270, 300, 478])
def test_uniform_to_htherm(temperature):
    si = IsingSpinIce(spin_angle=0, alpha=0, temperature=temperature)
    kB = 1.38064852e-23
    T = si.temperature
    m_therm = si.m_therm
    c = kB * T / m_therm
    attempt_freq = si.attempt_freq
    dt = si.therm_timescale

    def cdf_inv(x):
        return -c*np.log(np.log(x)/(-dt * attempt_freq))

    u = np.random.uniform(size=100)
    assert_array_equal(si._uniform_to_htherm(u), cdf_inv(u))

def test_htherm_magnitude():
    temperature = 300
    si = IsingSpinIce(size=(10,10), spin_angle=0, alpha=0, temperature=300, sw_b=1, sw_c=1, sw_beta=3, sw_gamma=3, hc=0.001, m_therm=860e3 * 1e-24)
    kB = 1.38064852e-23
    T = si.temperature
    m_therm = si.m_therm
    c = kB * T / m_therm
    attempt_freq = si.attempt_freq
    dt = si.therm_timescale

    si.set_random_seed(0x12340)
    si.update_thermal_noise()

    si.set_random_seed(0x12340)
    draw_uniform = si.rng.uniform(size = si.spin_count)
    h_len = si._uniform_to_htherm(draw_uniform)
    #h_therm = si._field_astroid_direction(h_len)

    assert_allclose(norm(si.h_therm, axis=1), h_len)

@pytest.mark.parametrize("magnitude", [0, 1/3, 1])
def test_field_astroid_direction_sw(magnitude):
    si = IsingSpinIce(size=(1,1), spin_angle=0, alpha=0, sw_b=1, sw_c=1, sw_beta=3, sw_gamma=3, hc=1)

    # Zero field should point at either 90+45 or 180+45 degrees
    # h_ext in q2
    si.set_h_ext([-1e-9, 1e-9])
    #magnitude = 1
    hdir = si._field_astroid_direction(magnitude)[0]
    expect = magnitude * np.array([np.cos(3*np.pi/4), np.sin(3*np.pi/4)])
    assert_allclose(hdir, expect, atol=1e-03, rtol=1e-03)

    # h_ext in q3
    si.set_h_ext([-1e-9, -1e-9])
    hdir = si._field_astroid_direction(magnitude)[0]
    expect = magnitude * np.array([np.cos(5*np.pi/4), np.sin(5*np.pi/4)])
    assert_allclose(hdir, expect, atol=1e-03, rtol=1e-03)

def test_field_astroid_direction_outside():
    si = IsingSpinIce(size=(1,1), spin_angle=0, alpha=0, sw_b=1, sw_c=1, sw_beta=3, sw_gamma=3, hc=1)

    # h_ext outside of astroid, should still point out
    si.set_h_ext([-.9, -.9])
    hdir = si._field_astroid_direction(1)[0]
    expect = np.array([np.cos(5*np.pi/4), np.sin(5*np.pi/4)])
    assert_allclose(hdir, expect, atol=1e-03, rtol=1e-03)

def test_field_astroid_direction_asw():
    # beta=gamma=2 and c=.5 gives an astroid with linear shape
    # where distance to astroid can be found analytically
    c = 0.5
    si = IsingSpinIce(size=(1,1), spin_angle=0, alpha=0, sw_b=1, sw_c=c, sw_beta=2, sw_gamma=2, hc=1)

    si.set_h_ext([0, 0.01])
    hdir = si._field_astroid_direction(1)[0]
    x = -2*c**2/(2+2*c**2)
    y = c * (x+1)
    expect = np.array([x, y])
    expect /= norm(expect)
    assert_allclose(hdir, expect, atol=1e-03, rtol=1e-03)

@pytest.mark.parametrize("angle", [0, 20, 30, 45, 60, 90, 112, 120, 135, 160, 180])
def test_field_astroid_direction_circle(angle):
    # Circular astroid should always yield direction radially along h_ext
    si = IsingSpinIce(size=(1,1), spin_angle=0, alpha=0, sw_b=1, sw_c=1, sw_beta=1, sw_gamma=1, hc=1)

    angle = np.deg2rad(angle)

    hpar = np.cos(np.pi/2+angle)
    hperp = np.sin(np.pi/2+angle)
    si.set_h_ext([.1*hpar, .1*hperp])
    hdir = si._field_astroid_direction(1)[0]
    expect = np.array([hpar, hperp])
    assert_allclose(hdir, expect, rtol=1e-03, atol=1e-03)

def test_field_astroid_direction_boundary():
    # Test when h_ext is exactly at the astroid boundary
    hc = 0.02
    #hc = 1
    b = 1
    c = 1
    beta = 3
    gamma = 3
    res = 101
    si = IsingSpinIce(size=(1,1), spin_angle=0, alpha=0, sw_b=b, sw_c=c, sw_beta=beta, sw_gamma=gamma, hc=hc, astroid_resolution=res)

    thetas = np.linspace(np.pi, 2*np.pi, res)

    h_perp = c * np.cos(thetas)
    h_par = np.sin(thetas)
    astroid = ASW(h_par, h_perp, b=b, c=c, beta=beta, gamma=gamma)

    for i in range(len(astroid)):
        h = np.array([hc * astroid[i], hc * h_perp[i]])
        si.set_h_ext(h)

        # Positive magnitude should always point out of astroid (towards switching)
        hdir = si._field_astroid_direction(1)[0]
        assert hdir[0] < 0, f"Failed for i={i} h={h/hc} {si.astroid_resolution} {astroid[i], h_perp[i]}"

        # Negative magnitude should always point into astroid (against switching)
        hdir = si._field_astroid_direction(-1)[0]
        assert hdir[0] > 0, f"Failed for i={i} h={h/hc}"

@pytest.mark.parametrize("magnitude", [1.01, 0.99])
def test_field_astroid_direction_flippable(magnitude):
    hc = 0.02
    b = 1
    c = 1
    beta = 3
    gamma = 3
    H = magnitude * hc
    res = 101
    si = IsingSpinIce(size=(1,1), spin_angle=0, alpha=0, sw_b=b, sw_c=c, sw_beta=beta, sw_gamma=gamma, hc=hc, astroid_resolution=res)

    assert len(si.flippable()) == 0

    dtheta = np.pi / 1024
    thetas = np.arange(np.pi + dtheta, 2*np.pi, dtheta)
    degs = np.rad2deg(thetas)

    h_perp = c * np.cos(thetas)
    h_par = np.sin(thetas)
    astroid = ASW(h_par, h_perp, b=b, c=c, beta=beta, gamma=gamma)

    for i in range(len(astroid)):
        h = np.array([H * astroid[i], H * h_perp[i]])

        si.set_h_ext(h)
        assert len(si.flippable()) == (1 if magnitude > 1 else 0)
        hdir = si._field_astroid_direction(hc)[0]
        assert hdir[0] < 0, f"Expected negative hdir for i={i} deg={degs[i]} h={h} hdir={hdir} h_ext={si.h_ext}"
        si.set_h_ext(h + hdir)
        assert len(si.flippable()) == 1, f"Should flip for i={i} deg={degs[i]} h={h} hdir={hdir} h_ext={si.h_ext}"

        si.set_h_ext(h)
        assert len(si.flippable()) == (1 if magnitude > 1 else 0)
        hdir = si._field_astroid_direction(-hc)[0]
        assert hdir[0] > 0, f"Expected positive hdir for i={i} deg={degs[i]} h={h} hdir={hdir} h_ext={si.h_ext}"
        si.set_h_ext(h + hdir)
        assert len(si.flippable()) == 0, f"Should not flip for i={i} theta={np.rad2deg(thetas[i])} h={h/hc} hdir={hdir} h_ext={si.h_ext/hc}"

def test_field_astroid_direction_negative():
    hc = 1
    si = IsingSpinIce(size=(1,1), spin_angle=0, alpha=0, sw_b=1, sw_c=1, sw_beta=3, sw_gamma=3, hc=hc)

    # External field is strong enough to switch
    phis = np.arange(90+1, 270-1, 1)
    rads = np.deg2rad(phis)
    H = 1.4 * hc
    hs = np.column_stack([H * np.cos(rads), H * np.sin(rads)])
    for phi, h in zip(phis, hs):
        si.set_h_ext(h)
        E = si.switching_energy()[0]
        assert len(si.flippable()) == 1
        assert E > 1e-10

        # A large negative thermal magnitude should prevent switching
        hdir = si._field_astroid_direction(-2)[0]
        si.set_h_ext(h + hdir)
        E = si.switching_energy()[0]
        assert len(si.flippable()) == 0, f"Flippable for phi={phi}"
        # Avoid numerical instability
        assert E < -1e-10, f"Numerical instability for phi={phi}"

def test_h_therm(opencl):
    si = IsingSpinIce(size=(10,10), spin_angle=0, temperature=300,
            hc=0.02, sw_b=1, sw_c=1, sw_beta=3, sw_gamma=3,
            m_therm=860e3 * 1e-23, therm_timescale=0.1, **opencl)

    assert si.temperature == 300

    h_therm1 = si.h_therm.copy()
    h_therm1_magnitude = norm(h_therm1, axis=1)
    n_flippable1 = len(si.flippable())
    assert n_flippable1 > 0

    assert_array_equal(si.total_fields(), h_therm1 + si.dipolar_fields())

    assert si.step() == 1

    assert_array_equal(si.h_therm, h_therm1)

    # Increase temperature
    si.set_temperature(320)
    assert si.temperature == 320

    for i in range(10):
        si.update_thermal_noise()
        h_therm2 = si.h_therm.copy()
        h_therm2_magnitude = norm(h_therm2, axis=1)
        n_flippable2 = len(si.flippable())

        assert not np.array_equal(h_therm1, h_therm2)
        # Magnitude of h_therm should be greater at higher temperatures
        assert np.sum(h_therm1_magnitude) < np.sum(h_therm2_magnitude)
        assert n_flippable1 < n_flippable2



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Dataset statistics.
"""
import numpy as np
import pandas as pd
from joblib import delayed
from flatspin.data import Dataset, read_table, save_table
from flatspin.cmdline import *

def load_stats(ds):
    df = read_table(ds.tablefile('stats'))
    col0, col1 = df.columns
    keys = list(df[col0])
    values = list(df[col1].astype(float))
    df = pd.DataFrame([values], columns=keys)
    return df

def main_stats(dataset, args):
    # Do it in parallell
    progress_bar = ProgressBar(desc="Stats", total=len(dataset))
    parallel = ParallelProgress(progress_bar, n_jobs=-1)

    stats = parallel(delayed(load_stats)(ds) for ds in dataset)

    progress_bar.close()

    df = pd.concat(stats, ignore_index=True)

    if args.output:
        out = pd.concat([dataset.index, df], axis=1)
        save_table(out, args.output)
    #df['runtime'] = pd.to_timedelta(df['runtime'], unit='s')

    stats = OrderedDict()
    for col, d in df.items():
        stats[col] = [ d.count(), d.mean(), d.std(), d.min(), d.max(), d.sum() ]
    stats = pd.DataFrame(stats, index=['count', 'mean', 'std', 'min', 'max', 'sum'])

    pd.options.display.float_format = '{:.1f}'.format
    print(stats)

# Parser needs to be accessible at module level for sphinxcontrib.autoprogram
parser = main_dataset_argparser(__doc__)

parser.add_argument('-q', '--stats', nargs='+', default=['runtime'],
        help='type of stats')

def main():
    args = parser.parse_args()

    ds = main_dataset(args)

    return main_stats(ds, args)

if __name__ == '__main__':
    main()



#!/usr/bin/env python3
import sys
import re
import numpy as np
import pandas as pd
import fnmatch
import itertools
import textwrap
from os import path
from joblib import delayed
import matplotlib.pyplot as plt

from flatspin.data import (Dataset, read_table, table_columns, col_group,
        match_column, match_columns, filter_df, save_table)

from flatspin.plotting import format_label, heatmap, format_labels
from flatspin.cmdline import (main_dataset_argparser, main_dataset,
        FilterAction, ProgressBar, ParallelProgress)

def connect_picker(fig, ax):
    leg = ax.get_legend()
    leg._line_highlight = -1
    #leg = axes[0].legend(loc='upper left', bbox_to_anchor=(1.0, 1.05))

    for i, (legline, legtext) in enumerate(zip(leg.get_lines(), leg.get_texts())):
        #line.set_picker(5) # 5 pts tolerance
        legline.set_picker(5)
        legtext.set_picker(5)
        legline._line_index = i
        legtext._line_index = i

    def onpick(event):
        if leg._line_highlight < 0:
            leg._line_highlight = event.artist._line_index
        else:
            leg._line_highlight = -1

        for j, line in enumerate(ax.get_lines()):
            if j == leg._line_highlight or leg._line_highlight < 0:
                line.set_alpha(1.0)
                line.set_zorder(10)
            else:
                line.set_alpha(0.1)
                line.set_zorder(0)
        fig.canvas.draw()

    fig.canvas.mpl_connect('pick_event', onpick)

def aslist(a):
    if a is None:
        return []

    if not isinstance(a, list):
        return [a]

    return a

def astuple(a):
    if a is not None and not isinstance(a, tuple):
        a = (a,)
    return a

def pivot(df, index=None, columns=None, values=None):
    """ Alternative pivot which supports multiple columns for index/columns """
    if index:
        df = df.set_index(index)

    if columns:
        df = df.set_index(columns, append=True).unstack(columns)

    if values:
        df = df[values]

    return df

def load_one_dataset_tables(dataset, patterns=None, key=None, columns=None, table_filter=None):
    """ Load tables from one dataset """
    assert len(dataset) == 1

    df = pd.DataFrame()
    for f in dataset.tablefiles(patterns):
        cols = []
        if columns:
            # Skip tables that have none of the columns we're interested in
            tablecols = table_columns(f)
            matches = match_columns(columns, tablecols)
            if not matches:
                continue
            cols.extend(list(set(matches)))

        d = read_table(f)

        if not df.columns.intersection(d.columns).empty:
            # Join on common columns
            df = df.merge(d, how='outer')
        else:
            # No common columns -> join on index
            df = df.join(d, how='outer')

    if table_filter:
        df = filter_df(df, **table_filter)

    return df

def load_dataset_tables(dataset, patterns=None, key=None, columns=None, table_filter=None):
    """ Load tables from datasets """
    # Load tables from each dataset in parallell
    progress_bar = ProgressBar(desc="Loading tables", total=len(dataset))
    parallel = ParallelProgress(progress_bar, n_jobs=-1)

    dfs = parallel(
        delayed(load_one_dataset_tables)(ds, patterns, key, columns, table_filter)
        for ds in dataset)

    progress_bar.close()

    # Concatenate tables vertically, adding a column for the key
    key = aslist(key)
    if key:
        keys = list(dataset.index[key].itertuples(index=False))
    else:
        keys = list(dataset.index.index)
        key = ['idx']

    df = pd.concat(dfs, sort=False, axis=0, keys=keys, names=key)
    df = df.reset_index(level=key) # trick to get key as a column

    return df

def load_file_table(filename, table_filter=None):
    """ Load one table from file """
    df = read_table(filename)
    if table_filter:
        df = filter_df(df, **table_filter)
    return df

def load_file_tables(files, table_filter=None):
    """ Load tables from file """
    dfs = [load_file_table(f, table_filter) for f in files]
    df = pd.concat(dfs, sort=False, axis=0)
    return df

def aggregate(df, x=None, y=None, groupby=None, agg='mean'):
    """ Perform aggregation across one or more axes """
    x = aslist(x)
    y = aslist(y)
    groupby = aslist(groupby)

    if x and y:
        # Aggregate over x + y
        df = df.groupby(x + y).agg(agg).reset_index()
    elif x:
        # Aggregate over x + groupby (optional)
        grp = list(set(groupby + x))
        df = df.groupby(grp).agg(agg).reset_index()
    elif groupby:
        # Aggregate over index + groupby
        # Move original index to its own column
        df = df.reset_index()
        idx_col = df.columns[0]
        # Group by index col + groupby
        df = df.groupby(groupby + [idx_col]).agg(agg).reset_index()
        # Restore index
        df.set_index(idx_col, inplace=True)
        df.index.name = None
    else:
        # Aggregate over index only
        df = df.groupby(level=0).agg(agg)

    return df

def plot_xy(df, x=None, y=None, groupby=None, agg=None, xlim=None, ylim=None, **kwargs):
    """ Plot x vs y as lines """
    y = aslist(y) if y else None
    groupby = aslist(groupby)

    agg_label = False
    if agg:
        orig_len = len(df)
        df = aggregate(df, x, None, groupby, agg)
        new_len = len(df)
        if new_len < orig_len:
            print(f"Aggregating ({agg}): {orig_len} rows -> {new_len} rows")
            agg_label = True

    if groupby and x not in groupby:
        df = pivot(df, columns=groupby)
        #df = df.set_index(groupby, append=True).unstack(groupby)
        if agg_label:
            df.columns.set_names(f'{agg}(y)', level=0, inplace=True)
        else:
            df.columns.set_names('y', level=0, inplace=True)

    fig, ax = plt.subplots()

    if isinstance(df.columns, pd.MultiIndex):
        levels = list(range(1, df.columns.nlevels))
        for key, d in df.groupby(level=levels, axis=1):
            d = d.dropna(how='all')
            key = astuple(key)
            xi = (x,) + key if x else None
            d.plot(x=xi, y=y, ax=ax, **kwargs)
        plt.xlabel(x)
    else:
        if agg_label:
            df = df.rename(columns={col: f"{agg}({col})" for col in df.columns if col != x})
            y = [f"{agg}({col})" for col in y] if y else None
        df.plot(x=x, y=y, ax=ax, **kwargs)

    if xlim:
        plt.xlim(*xlim)
    if ylim:
        plt.ylim(*ylim)

    connect_picker(fig, ax)

def plot_xyz(df, x=None, y=None, z=None, groupby=None, agg=None, xlim=None, ylim=None, zlim=None):
    """ Plot x vs y vs z as heatmap """
    x = aslist(x)
    y = aslist(y)
    z = aslist(z)
    groupby = aslist(groupby)

    zlabel = 'z'
    if len(z) == 1:
        zlabel = z[0]

    if agg:
        orig_len = len(df)
        df = aggregate(df, x, y, groupby, agg)
        new_len = len(df)
        if new_len < orig_len:
            print(f"Aggregating ({agg}): {orig_len} rows -> {new_len} rows")
            zlabel = f"{agg}({zlabel})"

    index = x[0] if len(x) == 1 else x if x else None
    columns = y[0] if len(y) == 1 else y if y else None
    values = z[0] if len(z) == 1 else z if z else None

    if groupby and not y:
        # Plot each column as a row in the heatmap
        columns = groupby
        df = pivot(df, index=index, columns=groupby)
    else:
        df = pivot(df, index=index, columns=columns, values=values)

    if isinstance(df.columns, pd.MultiIndex) and df.columns.names[0] is None:
        df.columns.set_names('y', level=0, inplace=True)

    xlabel = format_label(df.index.names)
    ylabel = format_label(df.columns.names)

    pcolor = heatmap(df.index, df.columns, df.T, xlabel, ylabel, zlabel, xlim, ylim, zlim)

    plt.colorbar(pcolor, label=zlabel)
    plt.xlabel(xlabel)
    plt.ylabel(ylabel)

def func_atan(x):
    return np.rad2deg(np.arctan2(x[1], x[0]))

funcs = {
    'norm': np.linalg.norm,
    'atan': func_atan,
}

def parse_func(column):
    """ Parse column function

    New columns may be specified in the form:
        name=function(col1 col2 col3)

    Returns a tuple (name, function, [col1, col2, col3])
    """
    m = re.match(r'(\w+)=(\w+)\((.*)\)', column)
    if m:
        name = m.group(1)
        fn = m.group(2)
        fn = funcs.get(fn, fn)
        cols = list(map(str.strip, m.group(3).split()))
        return (name, fn, cols)

    # Not a function (nop)
    return (column, None, [column])

def parse_funcs(columns):
    """ Parse column functions """
    names = []
    cols = []
    functions = []

    for c in columns:
        name, func, col = parse_func(c)
        names.append(name)
        cols.extend(col)
        if func:
            functions.append((name, func, col))

    return names, cols, functions

def main_plot(dataset, args):
    # Columns to plot
    x = None
    y = None
    z = None

    # Any function specifications
    # name=func(cols)
    # functions is a list of (name, func, cols) tuples
    functions = []

    # List of columns we need from the tables
    columns = []

    # Parse any functions for -x -y and -z
    if args.x:
        x, cols, funcs = parse_funcs(args.x)
        columns.extend(cols)
        functions.extend(funcs)

    if args.y:
        y, cols, funcs = parse_funcs(args.y)
        columns.extend(cols)
        functions.extend(funcs)

    if args.z:
        z, cols, funcs = parse_funcs(args.z)
        columns.extend(cols)
        functions.extend(funcs)

    #print(f"x={x} y={y} z={z} (args)")
    #print(f"columns={columns}")
    #print(f"functions={functions}")

    # Load tables
    patterns = [(p, p + '.*') for p in args.tables]
    patterns = list(itertools.chain(*patterns))

    dfs = []
    if dataset:
        df = load_dataset_tables(dataset, patterns, args.groupby, columns, args.table_filter)
        dfs.append(df)

    if args.files:
        df = load_file_tables(args.files, args.table_filter)
        dfs.append(df)

    df = pd.concat(dfs, sort=False, axis=0)

    # Round columns a bit
    df = df.round(10)

    # Apply functions as new columns
    # TODO: support for functions returning multiple values?
    for name, fn, cols in functions:
        cols = match_columns(cols, df.columns) or cols
        # ... or cols to cause KeyError when no match
        df[name] = df[cols].agg(fn, axis=1)

    # Now that we have the table, we can determine which columns to plot
    # (some column arguments may be patterns, e.g., "spin*x")
    if x:
        x = match_columns(x, df.columns) or x
    if y:
        y = match_columns(y, df.columns) or y
    if z:
        z = match_columns(z, df.columns) or z

    print(f"x={x} y={y} z={z}")

    groupby = args.groupby
    if dataset and not args.groupby:
        # Group by the dataset index by default
        groupby = 'idx'

    if z or y:
        # Keep only columns we're interested in
        xyz = aslist(x) + aslist(y) + aslist(z)
        df = df[list(set(xyz + aslist(groupby)))]

        # Drop rows where x, y and z are all NaN
        df = df.dropna(how='all', subset=xyz)

    if x:
        # Drop rows where x is NaN
        df = df.dropna(subset=x, how='any')
    if z and y:
        # Drop rows where y is NaN
        df = df.dropna(subset=y, how='any')

    print(df)

    if args.table_out:
        print("Saving table {}".format(args.table_out))
        save_table(df, args.table_out)

    if args.z:
        # 3D plot
        plot_xyz(df, x, y, z, groupby, args.aggregate, args.xlim, args.ylim, args.zlim)

    else:
        # 2D plot
        if x and len(x) > 1:
            raise ValueError("Multiple columns on x axis is not supported")
        x = x[0] if x else None

        plot_xy(df, x, y, groupby, args.aggregate, args.xlim, args.ylim,
                marker=args.marker, linestyle=args.linestyle)

    # Set the title
    title = args.title
    if dataset and title:
        params = dataset.params
        params.update(dataset.row(0))
        kw = {'name': dataset.name, 'basepath': dataset.basepath,
              'index': dataset.id(0), 'params': params}
        kw.update(params)
        kw.update(dataset.info)
        title = title.format(**kw)

    plt.title(title)

    if args.output:
        print("Saving figure {}".format(args.output))
        plt.savefig(args.output, dpi=args.dpi)
    else:
        plt.show()

    return 0

def main_list_cols(dataset, args):
    if dataset:
        for ds in dataset:
            patterns = [(p, p + '.*') for p in args.tables]
            patterns = list(itertools.chain(*patterns))
            files = ds.tablefiles(patterns)
            columns = list(map(table_columns, files))
            for filename, cols in zip(files, columns):
                print(f"{filename}:")
                print(f"  columns ({len(cols)}):")
                print(textwrap.fill(", ".join(cols), initial_indent="  ", subsequent_indent="  "))

    if args.files:
        columns = list(map(table_columns, args.files))
        for filename, cols in zip(args.files, columns):
            print(f"{filename}")
            print(f"  columns ({len(cols)}):")
            print(textwrap.fill(", ".join(cols), initial_indent="  ", subsequent_indent="  "))

    return 0


# Parser needs to be accessible at module level for sphinxcontrib.autoprogram
parser = main_dataset_argparser("Plot table data.")

parser.add_argument('-f', '--files', nargs='*',
        help='tables to load from file')
parser.add_argument('-g', '--tables', nargs='+', default=[],
        help='tables to load from the dataset')
parser.add_argument('-k', '--groupby', nargs='*',
        help='group by column(s)')
parser.add_argument('-v', '--list-cols', action='store_true',
        help='list available columns')
parser.add_argument('-x', nargs='*',
        help='name of column(s) to use for x axis')
parser.add_argument('-y', nargs='*',
        help='name of column(s) to use for y axis')
parser.add_argument('-z', nargs='*',
        help='name of column(s) to use for z axis')
parser.add_argument('-a', '--aggregate',
        help='aggregate method')
parser.add_argument('-c', '--table-filter', action=FilterAction, default={},
        help='select a subset of the table (column=value)')
parser.add_argument('--xlim', nargs=2, type=float, metavar=('XMIN', 'XMAX'))
parser.add_argument('--ylim', nargs=2, type=float, metavar=('YMIN', 'YMAX'))
parser.add_argument('--zlim', nargs=2, type=float, metavar=('ZMIN', 'ZMAX'))
parser.add_argument('--title', help='title format string')
parser.add_argument('--marker', help='marker style (XY plots only)')
parser.add_argument('--linestyle', help='line style (XY plot only)')
parser.add_argument('-to', '--table-out', help='dump result table to file')
parser.add_argument('--dpi', type=int, default=100,
        help='dots per inch (default: %(default)s)')

def main():
    args = parser.parse_args()

    # With -f, we won't load dataset tables unless explicitly told to with -b/--datapath
    ds = None
    if not args.files or args.datapath is not None:
        ds = main_dataset(args)

    if args.list_cols:
        return main_list_cols(ds, args)

    return main_plot(ds, args)

if __name__ == '__main__':
    main()



import numpy as np
import pandas as pd
import pytest
import collections

from flatspin.sweep import sweep

def test_sweep_spec():
    sweep_spec = {'a': '[1,2,3]'}
    sweep_params = list(sweep(sweep_spec))
    assert sweep_params == [(0, 0, {'a': 1}), (1, 0, {'a': 2}), (2, 0, {'a': 3})]

def test_sweep_arange():
    sweep_spec = {'a': 'np.arange(3)'}
    sweep_params = list(sweep(sweep_spec))
    assert sweep_params == [(0, 0, {'a': 0}), (1, 0, {'a': 1}), (2, 0, {'a': 2})]

def test_sweep_expression():
    sweep_spec = {'a': 'bar + np.arange(foo)'}
    params = {'foo': 3, 'bar': 42}
    sweep_params = list(sweep(sweep_spec, params=params))
    assert sweep_params == [(0, 0, {'a': 42}), (1, 0, {'a': 43}), (2, 0, {'a': 44})]

def test_sweep_multiple():
    sweep_spec = {'a': 'np.arange(3)', 'b': '[1,10]'}
    sweep_params = list(sweep(sweep_spec))
    assert sweep_params == [
            (0, 0, {'a': 0, 'b': 1}),
            (1, 0, {'a': 0, 'b': 10}),
            (2, 0, {'a': 1, 'b': 1}),
            (3, 0, {'a': 1, 'b': 10}),
            (4, 0, {'a': 2, 'b': 1}),
            (5, 0, {'a': 2, 'b': 10})]

def test_sweep_depend():
    sweep_spec = {'a': 'np.arange(3)', 'b': 'np.arange(a, 3)'}
    sweep_params = list(sweep(sweep_spec))
    assert sweep_params == [
            (0, 0, {'a': 0, 'b': 0}),
            (1, 0, {'a': 0, 'b': 1}),
            (2, 0, {'a': 0, 'b': 2}),
            (3, 0, {'a': 1, 'b': 1}),
            (4, 0, {'a': 1, 'b': 2}),
            (5, 0, {'a': 2, 'b': 2})]

    sweep_spec = {'a': 'np.arange(3)', 'b': 'np.arange(a)'}
    sweep_params = list(sweep(sweep_spec))
    assert sweep_params == [
            (0, 0, {'a': 0}),
            (1, 0, {'a': 1, 'b': 0}),
            (2, 0, {'a': 2, 'b': 0}),
            (3, 0, {'a': 2, 'b': 1})]

def test_sweep_repeat():
    sweep_spec = {'a': 'range(3)'}
    sweep_params = list(sweep(sweep_spec, repeat=2))
    seeds = [p['random_seed'] for i,j,p in sweep_params]
    assert sweep_params == [
            (0, 0, {'a': 0, 'repeat': 0, 'random_seed': seeds[0]}),
            (0, 1, {'a': 0, 'repeat': 1, 'random_seed': seeds[1]}),
            (1, 0, {'a': 1, 'repeat': 0, 'random_seed': seeds[0]}),
            (1, 1, {'a': 1, 'repeat': 1, 'random_seed': seeds[1]}),
            (2, 0, {'a': 2, 'repeat': 0, 'random_seed': seeds[0]}),
            (2, 1, {'a': 2, 'repeat': 1, 'random_seed': seeds[1]})]

def test_sweep_repeat_spec():
    sweep_spec = {'a': 'range(3)'}
    repeat_spec = {'b': 'range(2)'}
    sweep_params = list(sweep(sweep_spec, repeat_spec=repeat_spec))
    assert sweep_params == [
            (0, 0, {'a': 0, 'b': 0}), (0, 1, {'a': 0, 'b': 1}),
            (1, 0, {'a': 1, 'b': 0}), (1, 1, {'a': 1, 'b': 1}),
            (2, 0, {'a': 2, 'b': 0}), (2, 1, {'a': 2, 'b': 1})]

def test_sweep_repeat_specs():
    sweep_spec = {'a': 'range(3)'}
    repeat_spec = {'b': 'range(2)', 'c': '[10,20]'}
    sweep_params = list(sweep(sweep_spec, repeat_spec=repeat_spec))
    assert sweep_params == [
            (0, 0, {'a': 0, 'b': 0, 'c': 10}), (0, 1, {'a': 0, 'b': 1, 'c': 20}),
            (1, 0, {'a': 1, 'b': 0, 'c': 10}), (1, 1, {'a': 1, 'b': 1, 'c': 20}),
            (2, 0, {'a': 2, 'b': 0, 'c': 10}), (2, 1, {'a': 2, 'b': 1, 'c': 20})]

def test_sweep_repeat_seed():
    sweep_spec = {'a': '[42]'}
    seeds = []
    for i in range(10):
        sweep_params = sweep(sweep_spec, repeat=10, seed=0)
        seeds.append([p['random_seed'] for i,j,p in sweep_params])

    seeds = np.array(seeds)

    # Should get same sequence of seeds for each call to sweep()
    assert np.all(seeds == seeds[0])




import numpy as np
import pandas as pd
import pytest
from numpy.testing import assert_array_equal

from flatspin.data import *

def test_crop():
    # (height, width)
    X = [[1,2,3,4],
         [5,6,7,8],
         [9,1,2,3]]

    assert_array_equal(crop(X, ((1,1), (1,1))), [[6,7]])
    assert_array_equal(crop(X, (1,1)), [[6,7]])
    assert_array_equal(crop(X, 1), [[6,7]])
    assert_array_equal(crop(X, ((0,0), (1,1))),
            [[2,3],
             [6,7],
             [1,2]])
    assert_array_equal(crop(X, ((1,1), (0,0))), [[5,6,7,8]])

    # (time, height, width)
    X2 = [[[1,2,3,4],
           [5,6,7,8],
           [9,1,2,3]],
          [[4,5,6,7],
           [8,9,1,2],
           [3,4,5,6]]]

    assert_array_equal(crop(X2, ((1,1), (1,1))),
            [[[6,7]],
             [[9,1]]])
    assert_array_equal(crop(X2, ((0,0), (1,1))),
            [[[2,3],
              [6,7],
              [1,2]],
             [[5,6],
              [9,1],
              [4,5]]])
    assert_array_equal(crop(X2, ((1,1), (0,0))),
            [[[5,6,7,8]],
             [[8,9,1,2]]])


    # (time, height, width, 2) - vector image
    X3 = [[[[1,2],[3,4],[5,6]],
           [[7,8],[9,1],[2,3]],
           [[4,5],[6,7],[8,9]]],
          [[[0,1],[2,3],[4,5]],
           [[6,7],[8,0],[1,2]],
           [[3,4],[5,6],[7,8]]]]

    assert_array_equal(crop(X3, ((1,1), (1,1)), axis=-2),
            [[[[9,1]]],
             [[[8,0]]]])
    assert_array_equal(crop(X3, ((0,0), (1,1)), axis=-2),
            [[[[3,4]],
              [[9,1]],
              [[6,7]]],
             [[[2,3]],
              [[8,0]],
              [[5,6]]]])
    assert_array_equal(crop(X3, ((1,1), (0,0)), axis=-2),
            [[[[7,8],[9,1],[2,3]]],
             [[[6,7],[8,0],[1,2]]]])


def test_rolling_window():
    # (height, width)
    X = [[1,2,3,4],
         [5,6,7,8],
         [9,1,2,3]]

    Xr = rolling_window(X, 1)
    assert_array_equal(Xr, X)

    Xr = rolling_window(X, 2)
    expect = [[1+2+5+6, 2+3+6+7, 3+4+7+8],
              [5+6+9+1, 6+7+1+2, 7+8+2+3]]
    assert_array_equal(Xr, expect)

    Xr = rolling_window(X, (3,2))
    expect = [[1+2+5+6+9+1, 2+3+6+7+1+2, 3+4+7+8+2+3]]
    assert_array_equal(Xr, expect)

    Xr = rolling_window(X, (1,2))
    expect = [[1+2, 2+3, 3+4],
              [5+6, 6+7, 7+8],
              [9+1, 1+2, 2+3]]
    assert_array_equal(Xr, expect)

    Xr = rolling_window(X, 2, step=2)
    expect = [[1+2+5+6, 3+4+7+8]]
    assert_array_equal(Xr, expect)

    Xr = rolling_window(X, 2, step=(1,2))
    expect = [[1+2+5+6, 3+4+7+8],
              [5+6+9+1, 7+8+2+3]]
    assert_array_equal(Xr, expect)

    Xr = rolling_window(X, (1,2), step=(1,2))
    expect = [[1+2, 3+4],
              [5+6, 7+8],
              [9+1, 2+3]]
    assert_array_equal(Xr, expect)

    # (height, width, 2)
    X2 = [[[1,2],[3,4],[5,6],[7,8]],
          [[9,1],[2,3],[4,5],[6,7]]]

    Xr = rolling_window(X2, 1)
    assert_array_equal(Xr, X2)

    Xr = rolling_window(X2, (2,2,1), (2,2,1))
    expect = [[[1+3+9+2,2+4+1+3],[5+7+4+6,6+8+5+7]]]
    assert_array_equal(Xr, expect)

    # (time, height, width, 2)
    X3 = [[[[1,2],[3,4],[5,6],[7,8]],
           [[9,1],[2,3],[4,5],[6,7]]],
          [[[8,9],[1,2],[3,4],[5,6]],
           [[7,8],[9,1],[2,3],[4,5]]]]

    Xr = rolling_window(X3, 1)
    assert_array_equal(Xr, X3)

    Xr = rolling_window(X3, (1,2,2,1), (1,2,2,1))
    expect = [[[[1+3+9+2,2+4+1+3],[5+7+4+6,6+8+5+7]]],
              [[[8+1+7+9,9+2+8+1],[3+5+2+4,4+6+3+5]]]]
    assert_array_equal(Xr, expect)

    # Masked arrays
    X = [[1,2,3,4],
         [5,6,7,8],
         [9,1,2,3]]
    mask=[[0,1,0,1],
          [1,1,1,0],
          [1,1,0,1]]
    X = np.ma.array(X, mask=mask)

    Xr = rolling_window(X, 1)
    assert_array_equal(Xr, X)

    Xr = rolling_window(X, 2)
    expect = [[1, 3, 3+8],
              [0, 2, 8+2]]
    assert_array_equal(Xr, expect)

def test_archive_key():
    assert archive_key("foo.npz/table") == ("foo.npz", "table")
    assert archive_key("foo.npz") == ("foo.npz", None)

def test_get_format():
    assert get_format("table.csv") == "csv"
    assert get_format("table.npy") == "npy"
    assert get_format("table.npz") == "npz"
    assert get_format("table.hdf") == "hdf"

    assert get_format("out/table.csv") == "csv"
    assert get_format("out/table.npy") == "npy"
    assert get_format("out/table.npz") == "npz"
    assert get_format("out/table.hdf") == "hdf"

    assert get_format("out.npz/table") == "npz"
    assert get_format("out.hdf/table") == "hdf"

def test_is_archive():
    assert is_archive("foo.hdf")
    assert is_archive("foo.hdf/table")
    assert is_archive("foo.npz")
    assert is_archive("foo.npz/table")

    assert not is_archive("foo.out/table.csv")
    assert not is_archive("foo.out/table.npy")

def test_is_tablefile():
    assert is_tablefile("foo.out/table.csv")
    assert is_tablefile("foo.out/table.npy")

    assert is_tablefile("foo.hdf")
    assert is_tablefile("foo.hdf/table")
    assert is_tablefile("foo.npz/table.npz")
    assert is_tablefile("foo.npz")
    assert is_tablefile("foo.npz/table")
    assert is_tablefile("foo.npz/table.npz")

    assert not is_tablefile("foo.out/file.png")
    assert not is_tablefile("foo/file")

def test_listfiles_dir(tmp_path):
    filenames = [f"file{i}.csv" for i in range(10)]
    for f in filenames:
        open(os.path.join(tmp_path, f), 'w').close()

    expect = [os.path.join(tmp_path, f) for f in filenames]
    assert_array_equal(listfiles(tmp_path), expect)

def test_listfiles_npz(tmp_path):
    tables = {f"tab{i}": np.array([0]) for i in range(10)}
    filename = os.path.join(tmp_path, "data.npz")
    np.savez_compressed(filename, **tables)

    expect = [os.path.join(tmp_path, "data.npz", t) for t in tables.keys()]
    assert_array_equal(listfiles(filename), expect)

@pytest.mark.parametrize("fmt", table_formats)
def test_save_table(tmp_path, fmt):
    df = pd.DataFrame(np.random.randint(0, 10, (10,10),dtype='int64'),
            columns=[f"col{i}" for i in range(10)])

    filename = f"foo.{fmt}"
    if is_archive_format(fmt):
        filename = os.path.join(filename, 'table')

    filename = os.path.join(tmp_path, filename)

    save_table(df, filename)

    if is_archive_format(fmt):
        head, tail = os.path.split(filename)
        assert os.path.isfile(head)
    else:
        assert os.path.isfile(filename)

    df2 = read_table(filename)
    assert df.equals(df2)

    assert_array_equal(table_columns(filename), df.columns)

@pytest.mark.parametrize("fmt", table_formats)
def test_save_dict(tmp_path, fmt):
    data = {'foo': 1, 'bar': 'baz', 'arr': np.arange(3)}
    df = to_table(data)

    filename = f"foo.{fmt}"
    if is_archive_format(fmt):
        filename = os.path.join(filename, 'table')

    filename = os.path.join(tmp_path, filename)

    save_table(df, filename)

    if is_archive_format(fmt):
        head, tail = os.path.split(filename)
        assert os.path.isfile(head)
    else:
        assert os.path.isfile(filename)

    df2 = read_table(filename)

    assert_array_equal(np.array(df), np.array(df2))



#!/usr/bin/env python3
import os
import time
from joblib import delayed
from flatspin.data import *
from flatspin.cmdline import *

def convert_tables(outdir, tablefiles, data_format):
    if not is_archive(outdir):
        os.makedirs(outdir)

    for src in tablefiles:
        df = read_table(src)
        head, tail = os.path.split(src)
        name, ext = os.path.splitext(tail)

        if not is_archive(outdir):
            name = f"{name}.{data_format}"

        dst = os.path.join(outdir, name)
        save_table(df, dst)

def convert_dataset(dataset, destdir, fmt):
    if os.path.exists(destdir):
        # Refuse to overwrite an existing dataset
        raise FileExistsError(destdir)

    os.makedirs(destdir)

    queue = []
    keys = []
    new_outdir = []
    for i, tablefiles in enumerate(dataset.tablefiles(squash=False)):
        row = dataset.row(i)
        outdir = row['outdir']
        base, ext = os.path.splitext(outdir)
        if is_archive_format(fmt):
            outdir = f"{base}.{fmt}"
        else:
            outdir = f"{base}.out"
        new_outdir.append(outdir)

        dst = os.path.join(destdir, outdir)
        queue.append([dst, tablefiles, fmt])

    dataset.index['outdir'] = new_outdir
    dataset.save(destdir)

    queue = [delayed(convert_tables)(*q) for q in queue]

    progress_bar = ProgressBar(desc="dataset-convert", total=len(queue))
    parallel = ParallelProgress(progress_bar, n_jobs=1)
    parallel(queue)

    progress_bar.close()


# Parser needs to be accessible at module level for sphinxcontrib.autoprogram
parser = main_dataset_argparser("Convert between dataset formats.")

parser.add_argument('-f', '--format', choices=table_formats.keys(), default='npy',
        help='format of output files (default: %(default)s)')

def main():
    args = parser.parse_args()

    ds = main_dataset(args)

    convert_dataset(ds, args.output, args.format)

if __name__ == '__main__':
    main()



import numpy as np
import pytest
from .utils import approx, assert_allclose
from numpy.testing import assert_array_equal

from flatspin import SquareSpinIceOpen

def test_init(opencl):
    si = SquareSpinIceOpen(size=(5,4), **opencl)

    assert si.size == (5,4)
    assert si.spin_count == 5 * 4
    assert si.num_neighbors == 8

def test_labels(opencl):
    si = SquareSpinIceOpen(size=(2,4), **opencl)
    assert_array_equal(si.labels,
            [(0,0), (0,1),
             (1,0), (1,1),
             (2,0), (2,1),
             (3,0), (3,1)])

def test_indexof():
    si = SquareSpinIceOpen(size=(3,3))

    inds = [si.indexof(tuple(l)) for l in si.labels]
    assert_array_equal(inds, si.indices())

    # row 1
    assert_array_equal(si.L[1], [3, 4, 5])

    # col 1
    assert_array_equal(si.L[:,1], [1, 4, 7])

    # every other row
    assert_array_equal(si.L[::2], [0, 1, 2, 6, 7, 8])

    # every other column
    assert_array_equal(si.L[:,::2], [0, 2, 3, 5, 6, 8])

def test_geometry():
    si = SquareSpinIceOpen(size=(5,4))
    L = si.L

    # alternating 45 and 135 degrees
    a1 = np.pi/4
    a2 = a1 + np.pi/2

    assert_allclose(si.angle[L[0::2,0::2]], a1)
    assert_allclose(si.angle[L[1::2,1::2]], a1)
    assert_allclose(si.angle[L[0::2,1::2]], a2)
    assert_allclose(si.angle[L[1::2,0::2]], a2)

def test_num_neighbors():
    size = (10,10)

    # nearest neighborhood (default)
    si = SquareSpinIceOpen(size=size)
    assert si.num_neighbors == 8

    # global neighborhood
    si = SquareSpinIceOpen(size=size, neighbor_distance=np.inf)
    assert si.num_neighbors == si.spin_count - 1

def test_neighbors():
    size = (3,3)
    si = SquareSpinIceOpen(size=size)
    L = si.L

    # Lower left corner
    i = L[0,0]
    ns = list(si.neighbors(i))
    assert ns == [L[0,1], L[1,0], L[1,1]]

    # Lower center
    i = L[0,1]
    ns = list(si.neighbors(i))
    assert ns == [L[0,0], L[0,2], L[1,0], L[1,1], L[1,2]]

    # Lower right corner
    i = L[0,2]
    ns = list(si.neighbors(i))
    assert ns == [L[0,1], L[1,1], L[1,2]]

    # Left edge
    i = L[1,0]
    ns = list(si.neighbors(i))
    assert ns == [L[0,0], L[0,1], L[1,1], L[2,0], L[2,1]]

    # Center
    i = L[1,1]
    ns = list(si.neighbors(i))
    assert ns == [L[0,0], L[0,1], L[0,2], L[1,0], L[1,2], L[2,0], L[2,1], L[2,2]]

    # Right edge
    i = L[1,2]
    ns = list(si.neighbors(i))
    assert ns == [L[0,1], L[0,2], L[1,1], L[2,1], L[2,2]]

    # Top left corner
    i = L[2,0]
    ns = list(si.neighbors(i))
    assert ns == [L[1,0], L[1,1], L[2,1]]

    # Top center
    i = L[2,1]
    ns = list(si.neighbors(i))
    assert ns == [L[1,0], L[1,1], L[1,2], L[2,0], L[2,2]]

    # Top right
    i = L[2,2]
    ns = list(si.neighbors(i))
    assert ns == [L[1,1], L[1,2], L[2,1]]

def test_spin_dipolar_field(opencl):
    # default lattice spacing = 1 such that NN islands have coupling strength 1.5
    size = (3,3)
    si = SquareSpinIceOpen(size=size, **opencl)
    L = si.L

    j_NN = np.asarray([1.5, 0.5])
    j_L = np.asarray([1/np.sqrt(2), 0])
    j_T = np.asarray([1/(2*np.sqrt(2)), 0])


    assert si.spin_dipolar_field(L[1,1], L[0,1]) == approx(j_NN)
    assert si.spin_dipolar_field(L[1,1], L[1,0]) == approx(j_NN * [-1, 1])
    assert si.spin_dipolar_field(L[1,1], L[1,2]) == approx(j_NN * [-1, 1])
    assert si.spin_dipolar_field(L[1,1], L[2,1]) == approx(j_NN)

    assert si.spin_dipolar_field(L[1,1], L[0,0]) == approx(j_L)
    assert si.spin_dipolar_field(L[1,1], L[2,2]) == approx(j_L)

    assert si.spin_dipolar_field(L[1,1], L[0,2]) == approx(-j_T)
    assert si.spin_dipolar_field(L[1,1], L[2,0]) == approx(-j_T)

def test_dipolar_field(opencl):
    # default lattice spacing = 1 such that NN islands have coupling strength 1.5
    size = (4,4)
    si = SquareSpinIceOpen(size=size, alpha=1.0, **opencl)
    L = si.L

    j_NN = np.asarray([1.5, 0.5])
    j_L = np.asarray([1/np.sqrt(2), 0])
    j_T = np.asarray([1/(2*np.sqrt(2)), 0])

    assert si.dipolar_field(L[1,1]) == approx(2*j_L - 2*j_T + [0, 4] * j_NN)
    si.flip(L[1,1])
    assert si.dipolar_field(L[1,1]) == approx(-2*j_L + 2*j_T - [0, 4] * j_NN)
    si.flip(L[1,0])
    assert si.dipolar_field(L[1,1]) == approx(-2*j_L + 2*j_T - 2*j_NN)
    si.flip(L[1,2])
    assert si.dipolar_field(L[1,1]) == approx(-2*j_L + 2*j_T - [4, 0] *j_NN)

def test_external_field(opencl):
    a = 2*np.cos(np.pi/4)
    size = (4,4)
    si = SquareSpinIceOpen(size=size, lattice_spacing=a, **opencl)
    L = si.L

    assert si.external_field(L[0,0]) == approx([0, 0])
    assert si.external_field(L[1,0]) == approx([0, 0])

    cos45 = np.cos(np.pi/4)
    sin45 = np.sin(np.pi/4)

    # Horizontal
    si.set_h_ext((1,0))
    assert si.external_field(L[0,0]) == approx([cos45, -sin45])
    assert si.external_field(L[1,0]) == approx([-cos45, -sin45])

    # Vertical
    si.set_h_ext((0,1))
    assert si.external_field(L[0,0]) == approx([cos45, sin45])
    assert si.external_field(L[1,0]) == approx([cos45, -sin45])

    # 45 degrees
    si.set_h_ext((1,1))
    assert si.external_field(L[0,0]) == approx([2*cos45, 0])
    assert si.external_field(L[1,0]) == approx([0, -2*sin45])

def test_total_field(opencl):
    si = SquareSpinIceOpen(size=(4,4), **opencl)
    L = si.L
    si.randomize()

    i = L[2,2]
    assert tuple(si.total_field(i)) == tuple(si.dipolar_field(i) + si.external_field(i))

def test_dipolar_fields(opencl):
    si = SquareSpinIceOpen(size=(5,3), **opencl)
    si.randomize()

    h_dip = si.dipolar_fields()
    for i in si.indices():
        assert tuple(h_dip[i]) == tuple(si.dipolar_field(i))

def test_external_fields(opencl):
    si = SquareSpinIceOpen(size=(8,5), **opencl)
    si.randomize()

    # global external field
    si.set_h_ext((-.2,.1))

    h_ext = si.external_fields()
    for i in si.indices():
        assert tuple(h_ext[i]) == approx(tuple(si.external_field(i)))

    # local external field
    h_ext = np.random.uniform(-10, 10, (si.spin_count, 2))
    si.set_h_ext(h_ext)

    h_ext = si.external_fields()
    for i in si.indices():
        assert tuple(h_ext[i]) == approx(tuple(si.external_field(i)))

def test_total_fields(opencl):
    si = SquareSpinIceOpen(size=(4,4), **opencl)
    si.randomize()

    h_tot = si.total_fields()
    for i in si.indices():
        assert tuple(h_tot[i]) == tuple(si.total_field(i))

@pytest.mark.parametrize("switching", ["budrikis", "sw"])
def test_flippable(opencl, switching):
    si = SquareSpinIceOpen(size=(4,4), alpha=1.0, neighbor_distance=np.inf, switching=switching, hc=1.2, **opencl)
    L = si.L

    expected = 0
    if switching == 'sw':
        expected = 4

    assert len(si.flippable()) == expected

    si.flip(L[1,0])
    E = si.switching_energy()

    flippable = [i for i in si.indices() if E[i] > 0]

    assert_array_equal(si.flippable(), flippable)

    si.polarize()
    si.set_h_ext([0, -10])

    assert len(si.flippable()) == si.spin_count

@pytest.mark.parametrize("switching", ["budrikis", "sw"])
def test_alpha(opencl, switching):
    hc = 1.0
    h_ext = -(hc + 1.1)
    si = SquareSpinIceOpen(size=(4,4), alpha=1.0, neighbor_distance=np.inf, switching=switching, hc=hc, h_ext=(h_ext, 0), **opencl)
    h_dip = si.dipolar_fields()
    h_external = si.external_fields()
    h_tot = si.total_fields()
    n_flip = len(si.flippable())

    assert 0 < n_flip < si.spin_count

    # adjust alpha only
    alpha = 0.5
    si = SquareSpinIceOpen(size=(4,4), neighbor_distance=np.inf, switching=switching, hc=hc, h_ext=(h_ext, 0), alpha=alpha, **opencl)

    assert np.all(alpha * h_dip == si.dipolar_fields())
    assert np.all(h_external == si.external_fields())
    assert_allclose(alpha * h_dip + h_external, si.total_fields())

    # adjust alpha, rescale hc and h_ext
    alpha = 0.5
    hc *= alpha
    h_ext *= alpha
    si = SquareSpinIceOpen(size=(4,4), neighbor_distance=np.inf, switching=switching, hc=hc, h_ext=(h_ext, 0), alpha=alpha, **opencl)

    assert np.all(alpha * h_dip == si.dipolar_fields())
    assert np.all(alpha * h_external == si.external_fields())
    assert np.all(alpha * h_tot == si.total_fields())
    assert len(si.flippable()) == n_flip

@pytest.mark.parametrize("switching", ["budrikis", "sw"])
def test_relax(opencl, switching):
    si = SquareSpinIceOpen(size=(4,4), alpha=1.0, neighbor_distance=np.inf, switching=switching, hc=1.0, **opencl)
    si.flip(si.L[1,1])
    assert si.relax() > 0
    assert len(si.flippable()) == 0

def test_energy(opencl):
    si = SquareSpinIceOpen(size=(4,4), hc=1.0, **opencl)
    E0 = si.energy()
    si.flip(si.L[0,0])
    E1 = si.energy()
    assert E1[0] > E0[0]

@pytest.mark.parametrize("switching", ["budrikis", "sw"])
def test_total_energy(opencl, switching):
    si = SquareSpinIceOpen(size=(6,5), neighbor_distance=np.inf, switching=switching, hc=1.0, **opencl)

    E = si.total_energy()
    si.flip(si.L[0,0])
    assert si.total_energy() > E

    si.randomize()
    E = si.total_energy()
    while si.step():
        assert si.total_energy() < E
        E = si.total_energy()

def test_find_vertices():
    si = SquareSpinIceOpen(size=(4,4))
    L = si.L
    vi, vj, indices = si.find_vertices()

    assert_array_equal(vi, [0, 0, 1, 2, 2])
    assert_array_equal(vj, [0, 2, 1, 0, 2])

    assert tuple(indices[0]) == (L[0,0], L[0,1], L[1,0], L[1,1])
    assert tuple(indices[1]) == (L[0,2], L[0,3], L[1,2], L[1,3])

    assert tuple(indices[2]) == (L[1,1], L[1,2], L[2,1], L[2,2])

    assert tuple(indices[3]) == (L[2,0], L[2,1], L[3,0], L[3,1])
    assert tuple(indices[4]) == (L[2,2], L[2,3], L[3,2], L[3,3])

def test_vertex_type():
    si = SquareSpinIceOpen(size=(4,4))
    spin = [1, -1, 1, 1,
            1, -1, 1, 1,
            1, 1, 1, -1,
            -1, -1, 1, 1]
    si.set_spin(spin)

    vertex = si.vertices()

    assert si.vertex_type(vertex[0]) == 1
    assert si.vertex_type(vertex[1]) == 2
    assert si.vertex_type(vertex[2]) == 3
    assert si.vertex_type(vertex[3]) == 4
    assert si.vertex_type(vertex[4]) == 3

def test_spin_grid(opencl):
    ls = 1/3
    si = SquareSpinIceOpen(size=(4,4), lattice_spacing=ls, **opencl)
    spin = [1, 1, 1, 1,
            1, -1, 1, 1,
            1, 1, 1, -1,
            -1, -1, 1, 1]
    si.set_spin(spin)

    # grid_spacing == lattice_spacing
    grid = si.spin_grid()
    assert grid.shape == (4, 4, 2)
    expect = [[[1,1],  [-1,1],  [1,1],  [-1,1]],
              [[-1,1], [-1,-1], [-1,1], [1,1]],
              [[1,1],  [-1,1],  [1,1],  [1,-1]],
              [[1,-1], [-1,-1], [-1,1], [1,1]]]
    assert_allclose(grid, np.array(expect)*np.cos(np.pi/4))

    # grid_spacing < lattice_spacing
    grid = si.spin_grid(ls / 2)
    assert grid.shape == (7, 7, 2)
    expect = [[[1,1],  [0,0], [-1,1],  [0,0], [1,1],  [0,0], [-1,1]],
              [[0,0],  [0,0], [0,0],   [0,0], [0,0],  [0,0], [0,0]],
              [[-1,1], [0,0], [-1,-1], [0,0], [-1,1], [0,0], [1,1]],
              [[0,0],  [0,0], [0,0],   [0,0], [0,0],  [0,0], [0,0]],
              [[1,1],  [0,0], [-1,1],  [0,0], [1,1],  [0,0], [1,-1]],
              [[0,0],  [0,0], [0,0],   [0,0], [0,0],  [0,0], [0,0]],
              [[1,-1], [0,0], [-1,-1], [0,0], [-1,1], [0,0], [1,1]]]
    assert_allclose(grid, np.array(expect)*np.cos(np.pi/4))

    # grid_spacing > lattice_spacing
    grid = si.spin_grid(2 * ls)
    assert grid.shape == (2, 2, 2)
    expect = [[[-2,2], [0,4]],
              [[0,0], [2,2]]]
    assert_allclose(grid, np.array(expect)*np.cos(np.pi/4))

    # sum of all spins
    grid = si.spin_grid(si.spin_count * ls)
    assert grid.shape == (1, 1, 2)
    expect = [[[0,8]]]
    assert_allclose(grid, np.array(expect)*np.cos(np.pi/4))

def test_flip_mode_max(opencl):
    # With flip_mode=max, the flip order should be the same regardless of the
    # sample resolution of the external field, as long as the field has the
    # same extremes (min/max values)
    def do_test(ts):
        si = SquareSpinIceOpen(size=(4,4), disorder=0.05, neighbor_distance=np.inf,
                flip_mode='max', **opencl)

        H = si.hc/np.cos(np.pi/4)
        t = np.linspace(0, 2*np.pi, ts, endpoint=False)
        h_ext = H*np.sin(t)
        assert max(h_ext) == H

        si.polarize()
        for h in h_ext:
            si.set_h_ext((0, h))
            si.relax()
        return si.spin

    spin1000 = do_test(1024)
    spin100 = do_test(128)
    assert_array_equal(spin100, spin1000)
